# pylint: disable=unused-import
from copy import deepcopy
from functools import partial

from yaml import load
try:
    from yaml import CSafeLoader as Loader
except ImportError:  # nocv
    from yaml import SafeLoader as Loader
from django.db.models.query import ModelIterable
from django.db.models.fields import CharField, TextField, IntegerField, BooleanField, AutoField    # noqa: F401

from .models import BQuerySet, BaseModel
from .models.base import ModelBaseClass
from .utils import raise_context
from .tools import get_file_value, multikeysort  # pylint: disable=import-error


class Query(dict):
    __slots__ = 'queryset', 'combinator', 'is_sliced', 'select_for_update', 'select_related', 'empty'
    distinct_fields = False

    def __init__(self, queryset, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queryset = queryset
        self.combinator = None
        self.is_sliced = False
        self.select_for_update = False
        self.select_related = False
        self['standard_ordering'] = True

    @property
    def model(self):
        return self.queryset.model

    @property
    def standard_ordering(self):
        return self['standard_ordering']

    @standard_ordering.setter
    def standard_ordering(self, value):
        self['standard_ordering'] = bool(value)

    def chain(self):
        return self.clone()

    def clone(self):
        query = deepcopy(self)
        if 'custom_queryset_kwargs' in self:
            query['custom_queryset_kwargs'] = self['custom_queryset_kwargs']
        return query

    def _check_data(self, check_type, data):
        # pylint: disable=protected-access
        if getattr(self, 'empty', False):
            return False
        check_data = self.get(check_type, {})
        if check_type == 'exclude' and not check_data:
            return False
        meta = self.model._meta
        for filter_name, filter_data in check_data.items():
            filter_name = filter_name.replace('__exact', '')
            filter_name__cleared = filter_name.split('__')[0]
            if filter_name__cleared == 'pk':
                filter_name = '__'.join(
                    [meta.pk.attname] + filter_name.split('__')[1:]
                )
                filter_name__cleared = meta.pk.attname
            try:
                value = data[filter_name__cleared]
            except KeyError:
                continue
            field = meta._forward_fields_map[filter_name__cleared]
            if isinstance(filter_data, (list, tuple, set)):
                filter_data = map(field.to_python, filter_data)
            else:
                filter_data = field.to_python(filter_data)
            if '__in' in filter_name and value not in filter_data:
                return False
            elif '__in' not in filter_name and value != filter_data:
                return False
        return True

    def check_in_query(self, data):
        return self._check_data('filter', data) and not self._check_data('exclude', data)

    def set_empty(self):
        self.empty = True

    def set_limits(self, low: int = None, high: int = None):
        self['low_mark'], self['high_mark'] = low, high
        self.is_sliced = True

    def has_results(self, *args, **kwargs):
        # pylint: disable=unused-argument
        return bool(self.queryset.all()[:2])

    def get_count(self, using):
        # pylint: disable=unused-argument
        return len(self.queryset.all())

    def can_filter(self):
        return self.get('low_mark', None) is None and self.get('high_mark', None) is None

    def clear_ordering(self, *args, **kwargs):
        # pylint: disable=unused-argument
        self['ordering'] = []

    def add_ordering(self, *ordering):
        self['ordering'] = ordering

    @property
    def order_by(self):
        return self.get('ordering', ())


class CustomModelIterable(ModelIterable):
    __slots__ = ()

    def values_handler(self, unit):
        # pylint: disable=no-member
        return {f: unit.get(f) for f in self.fields}

    def construct_instance(self, data, model, only_fields, defer_fields):
        if only_fields is not None:
            data = {k: v for k, v in data.items() if k in only_fields}
        elif defer_fields is not None:
            data = {k: v for k, v in data.items() if k not in defer_fields}
        for field in model._meta.get_fields():
            if field.attname in data:
                with raise_context():
                    data[field.attname] = field.to_python(data[field.attname])
        return model(**data)

    def __iter__(self):
        # pylint: disable=protected-access
        queryset = self.queryset
        model = queryset.model
        query = queryset.query
        model_data = model._get_data(
            chunked_fetch=self.chunked_fetch,
            **query.get('custom_queryset_kwargs', {})
        )
        if isinstance(model._meta.pk, AutoField):
            for idx, item in enumerate(model_data, 1):
                item[model._meta.pk.attname] = idx
        model_data = list(filter(query.check_in_query, model_data))
        ordering = query.order_by
        if ordering:
            ordering = list(ordering)
            for idx, value in enumerate(ordering):
                if value in ('pk', '-pk'):
                    ordering[idx] = value.replace('pk', model._meta.pk.name)
            model_data = multikeysort(
                model_data,
                ordering,
                not query.standard_ordering
            )
        elif not query.standard_ordering:
            model_data.reverse()
        low = query.get('low_mark', 0)
        high = query.get('high_mark', len(model_data))
        fields = getattr(self, 'fields', None)
        if fields is None:
            handler = partial(
                self.construct_instance,
                model=model,
                only_fields=getattr(self, 'only_fields', None),
                defer_fields=getattr(self, 'defer_fields', None),
            )
        else:
            handler = self.values_handler
        for data in model_data[low:high]:
            yield handler(data)


class CustomQuerySet(BQuerySet):
    __slots__ = ()
    custom_iterable_class = CustomModelIterable
    custom_query_class = Query

    def _filter_or_exclude(self, negate, *args, **kwargs):
        clone = self._chain()
        if negate:
            filter_type = 'exclude'
        else:
            filter_type = 'filter'
        clone.query[filter_type] = clone.query.get(filter_type, {})
        clone.query[filter_type].update(kwargs)
        for q_arg in filter(lambda x: isinstance(x, dict), filter(bool, args)):
            clone.query[filter_type].update(q_arg)  # nocv
        return clone

    _filter_or_exclude_inplace = _filter_or_exclude

    def last(self):
        return self.reverse().first()

    def first(self):
        return next(iter(self), None)

    def values(self, *fields, **expressions):
        assert not expressions, 'Expressions is not supported on custom non-database models.'
        clone = self._clone()
        clone.__iterable_class__ = type('CustomModelIterableValues', (CustomModelIterable,), {'fields': fields})
        return clone

    def setup_custom_queryset_kwargs(self, **kwargs):
        qs = self._chain()
        qs.query['custom_queryset_kwargs'] = kwargs
        return qs

    def only(self, *fields):
        clone = self._clone()
        clone.__iterable_class__ = type('CustomModelIterable', (CustomModelIterable,), {'only_fields': fields})
        return clone

    def defer(self, *fields):
        clone = self._clone()
        clone.__iterable_class__ = type('CustomModelIterable', (CustomModelIterable,), {'defer_fields': fields})
        return clone


class CustomModelBase(ModelBaseClass):
    def __new__(mcs, name, bases, attrs, **kwargs):
        new_class = super(CustomModelBase, mcs).__new__(mcs, name, bases, attrs, **kwargs)
        if not new_class._meta.abstract:
            pk_name = new_class._meta.pk.attname
            new_class.add_to_class(
                pk_name,
                property(new_class.get_pk_value, new_class.set_pk_value)
            )
        return new_class


class ListModel(BaseModel, metaclass=CustomModelBase):
    """
    Custom model which uses a list of dicts with data (attribute `ListModel.data`) instead of database records.
    Useful when you have a simple list of data.

    Examples:
        .. sourcecode:: python

            from vstutils.custom_model import ListModel, CharField


            class Authors(ListModel):
                name = CharField(max_length=512)

                data = [
                    {"name": "Sergey Klyuykov"},
                    {"name": "Michael Taran"},
                ]

    Sometimes, it may be necessary to switch the data source. For these purposes,
    you should use the `setup_custom_queryset_kwargs` function, which takes various named arguments,
    which are also passed to the data initialization function.
    One such argument for :class:`ListModel` is date_source, which takes any iterable object.

    Examples:
        .. sourcecode:: python

            from vstutils.custom_model import ListModel, CharField


            class Authors(ListModel):
                name = CharField(max_length=512)

            qs = Authors.objects.setup_custom_queryset_kwargs(data_source=[
                {"name": "Sergey Klyuykov"},
                {"name": "Michael Taran"},
            ])

    In this case, we setup source list via `setup_custom_queryset_kwargs` function, and any other chained call
    is going to work with this data.
    """

    #: List with data dicts. Empty by default.
    data = []
    objects = CustomQuerySet.as_manager()

    class Meta:
        abstract = True

    def get_pk_value(self):
        return getattr(
            self,
            f'_{self.__class__._meta.pk.attname}',
            None
        )

    def set_pk_value(self, value):
        setattr(
            self,
            f'_{self.__class__._meta.pk.attname}',
            value
        )

    @classmethod
    def _get_data(cls, chunked_fetch=False, data_source=None):
        # pylint: disable=unused-argument
        return deepcopy(cls.data if data_source is None else data_source)


class FileModel(ListModel):
    """
    Custom model which loads data from YAML-file instead of database.
    Path to the file stored in `FileModel.file_path` attribute.


    Examples:
        Source file stored in `/etc/authors.yaml` with content:

        .. sourcecode:: YAML

            - name: "Sergey Klyuykov"
            - name: "Michael Taran"

        Example:

        .. sourcecode:: python

            from vstutils.custom_model import FileModel, CharField


            class Authors(FileModel):
                name = CharField(max_length=512)

                file_path = '/etc/authors.yaml'

    """

    class Meta:
        abstract = True

    @classmethod
    def load_file_data(cls):
        # pylint: disable=no-member
        return get_file_value(cls.file_path, strip=False)

    @classmethod
    def _get_data(cls, chunked_fetch=False):
        return load(cls.load_file_data(), Loader=Loader)



import functools
import operator

import ormsgpack
from django.utils.functional import Promise, LazyObject
from rest_framework.renderers import BaseRenderer
from rest_framework.settings import api_settings
from drf_orjson_renderer.renderers import ORJSONRenderer as BaseORJSONRenderer  # type: ignore

from ..utils import get_if_lazy


class ORJSONRenderer(BaseORJSONRenderer):
    @staticmethod
    def default(obj):
        # pylint: disable=protected-access
        if isinstance(obj, Promise):
            obj = obj._proxy____cast()
        elif isinstance(obj, LazyObject):  # nocv
            obj = get_if_lazy(obj)
        return BaseORJSONRenderer.default(obj)

    def render(self, data, media_type=None, renderer_context=None):
        if renderer_context and getattr(renderer_context['request'], 'is_bulk', False):
            return data
        return super().render(data, media_type, renderer_context)


class MsgpackRenderer(BaseRenderer):
    """
    Renderer which serializes to MessagePack.
    """

    media_type = 'application/msgpack'
    format = 'msgpack'
    render_style = 'binary'
    charset = None
    options = functools.reduce(
        operator.or_,
        api_settings.user_settings.get("MSGPACK_RENDERER_OPTIONS", ()),
        ormsgpack.OPT_SERIALIZE_NUMPY,
    )

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Renders *obj* into serialized MessagePack.
        """
        if data is None:
            return ''  # nocv
        return ormsgpack.packb(data, default=ORJSONRenderer.default, option=self.options)



"""
Additional serializer fields for generating OpenAPI and GUI.
"""
# pylint: disable=too-many-lines
import logging
import typing as _t
import copy
import json
import functools
import base64
from urllib.parse import quote

import orjson
from rest_framework.serializers import CharField, IntegerField, FloatField, ModelSerializer, Serializer
from rest_framework.fields import empty, SkipField, get_error_detail, Field
from rest_framework.exceptions import ValidationError
from django.apps import apps
from django.db import models
from django.utils.html import escape, strip_tags
from django.utils.functional import SimpleLazyObject, lazy, cached_property
from django.core.exceptions import ValidationError as DjangoValidationError
from django.core.files.uploadedfile import SimpleUploadedFile
from django.db.models.fields.files import FieldFile

from ..models import fields as vst_model_fields
from ..utils import raise_context, get_if_lazy, raise_context_decorator_with_default, translate, lazy_translate

DependenceType = _t.Optional[_t.Dict[_t.Text, _t.Text]]
DEFAULT_NAMED_FILE_DATA = {"name": None, "content": None, 'mediaType': None}


class VSTCharField(CharField):
    """
    CharField (extends :class:`rest_framework.fields.CharField`).
    This field translate any json type to string for model.
    """

    __slots__ = ()

    def to_internal_value(self, data) -> _t.Text:
        with raise_context():
            if not isinstance(data, str):
                data = json.dumps(data)
        data = str(data)
        return super().to_internal_value(data)


class FileInStringField(VSTCharField):
    """
    Field extends :class:`.VSTCharField` and saves file's content as string.

    Value must be text (not binary) and saves in model as is.

    .. note::
        Take effect only in GUI. In API it would behave as :class:`.VSTCharField`.
    """

    __slots__ = ()


class SecretFileInString(FileInStringField):
    """
    Field extends :class:`.FileInStringField`, but hides it's value in admin interface.

    Value must be text (not binary) and saves in model as is.

    .. note::
        Take effect only in GUI. In API it would behave as :class:`.VSTCharField`.
    """

    __slots__ = ()

    def __init__(self, **kwargs):
        kwargs['style'] = {'input_type': 'password'}
        super().__init__(**kwargs)


class BinFileInStringField(FileInStringField):
    """
    Field extends :class:`.FileInStringField`, but work with binary(base64) files.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = ()


class CSVFileField(FileInStringField):
    """
    Field extends :class:`.FileInStringField`, using for works with csv files.
    This field provides the display of the loaded data in the form of a table.

    :param items: The config of the table. This is a drf or vst serializer which includes char fields
                  which are the keys in the dictionaries into which the data from csv is serialized
                  and the  names for columns in a table.
                  The fields must be in the order you want them to appear in the table.
                  Following options may be included:

                  - ``label``: human readable column name
                  - ``required``: Defines whether the field should be required. False by default.
    :type items: Serializer

    :param min_column_width: Minimum cell width. Default is 200 px.
    :type min_column_width: int
    """
    __slots__ = ('min_column_width', 'items', 'delimiter')

    items: Serializer
    min_column_width: int
    delimiter: _t.Text

    def __init__(self, items=None, min_column_width=200, delimiter=',', **kwargs):
        self.items = items
        self.min_column_width = min_column_width
        self.delimiter = delimiter
        super().__init__(**kwargs)


class AutoCompletionField(VSTCharField):
    """
    Field that provides autocompletion on frontend, using specified list of objects.

    :param autocomplete: Autocompletion reference. You can set list/tuple with
                         values or set OpenApi schema definition name.
                         For definition name GUI will find optimal link and
                         will show values based on ``autocomplete_property`` and
                         ``autocomplete_represent`` arguments.
    :type autocomplete: list,tuple,str
    :param autocomplete_property: this argument indicates which attribute will be
                                  get from OpenApi schema definition model as value.
    :type autocomplete_property: str
    :param autocomplete_represent: this argument indicates which attribute will be
                                   get from OpenApi schema definition model as represent value.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``True``.
    :type use_prefetch: bool

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """
    __slots__ = 'autocomplete', 'autocomplete_property', 'autocomplete_represent', 'use_prefetch'

    autocomplete: _t.Text
    autocomplete_property: _t.Text
    autocomplete_represent: _t.Text
    use_prefetch: bool

    def __init__(self, **kwargs):
        self.autocomplete = kwargs.pop('autocomplete')
        self.autocomplete_property = None  # type: ignore
        if not isinstance(self.autocomplete, (list, tuple)):
            self.autocomplete_property = kwargs.pop('autocomplete_property', 'id')
            self.autocomplete_represent = kwargs.pop('autocomplete_represent', 'name')
            self.use_prefetch = kwargs.pop('use_prefetch', True)
        super().__init__(**kwargs)


class CommaMultiSelect(VSTCharField):
    """
    Field containing a list of values with specified separator(default: ",").
    Gets list of values from another model or custom list. Provides autocompletion as :class:`.AutoCompletionField`,
    but with comma-lists.
    Suited for property-fields in model where main logic is already implemented or
    with :class:`CharField`.

    :param select: OpenApi schema definition name or list with values.
    :type select: str,tuple,list
    :param select_separator: separator of values. Default is comma.
    :type select_separator: str
    :param select_property,select_represent: work as ``autocomplete_property`` and ``autocomplete_represent``.
                                             Default is ``name``.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``False``.
    :param make_link: Show value as link to model. Default is ``True``.
    :param dependence: Dictionary, where keys are name of field from the same model, and values are name of query filter
                       .If at least one of the fields that we depend on is non nullable, required and set to null,
                       autocompletion list will be empty and field will be disabled.
    :type dependence: dict


    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = (
        'select_model',
        'select_separator',
        'select_property',
        'select_represent',
        'use_prefetch',
        'make_link',
        'dependence'
    )

    select_model: _t.Text
    select_separator: _t.Text
    select_property: _t.Text
    select_represent: _t.Text
    use_prefetch: bool
    make_link: bool
    dependence: DependenceType

    def __init__(self, **kwargs):
        self.select_model = kwargs.pop('select')
        self.select_separator = kwargs.pop('select_separator', ',')
        self.select_property = None  # type: ignore
        if not isinstance(self.select_model, (list, tuple)):
            self.select_property = kwargs.pop('select_property', 'name')
            self.select_represent = kwargs.pop('select_represent', 'name')
        self.use_prefetch = kwargs.pop('use_prefetch', False)
        self.make_link = kwargs.pop('make_link', True)
        self.dependence = kwargs.pop('dependence', None)
        super().__init__(**kwargs)

    def to_internal_value(self, data: _t.Union[_t.Text, _t.Sequence]) -> _t.Text:
        return self.to_representation(data)  # nocv

    def to_representation(self, value: _t.Union[_t.Text, _t.Sequence, _t.Iterator]) -> _t.Text:
        if isinstance(value, str):
            value = map(str, filter(bool, value.split(self.select_separator)))
        return self.select_separator.join(value)


class DynamicJsonTypeField(VSTCharField):
    """
    Field which type is based on another field. It converts value to internal string
    and represent field as json object.

    :param field: field in model which value change will change type of current value.
    :type field: str
    :param types: key-value mapping where key is value of subscribed field and
                  value is type (in OpenApi format) of current field.
    :type type: dict
    :param choices: variants of choices for different subscribed field values.
                    Uses mapping where key is value of subscribed field and
                    value is list with values to choice.
    :type choices: dict


    .. note::
        Effective only in GUI. In API works similar to :class:`.VSTCharField` without value modifications.
    """
    __slots__ = 'field', 'choices', 'types'

    field: _t.Text
    choices: _t.Dict
    types: _t.Dict

    def is_json(self, real_field):
        return isinstance(real_field, Serializer)

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field')
        self.choices = kwargs.pop('choices', {})
        self.types = kwargs.pop('types', {})
        super().__init__(**kwargs)

    def get_attribute(self, instance):
        return instance

    @raise_context_decorator_with_default(default=None)
    def get_real_field(self, data) -> _t.Optional[Field]:
        if isinstance(data, dict):
            depend_value = data.get(self.field, None)
        else:
            depend_value = getattr(data, self.field, None)
        field = self.types.get(depend_value)
        if isinstance(field, Field):
            return field
        return None

    def to_internal_value(self, data):
        real_field = self.get_real_field({
            self.field: getattr(self.parent.instance, self.field, None),
            **self.parent.initial_data,
        })
        if real_field:
            data = real_field.to_internal_value(data)
        if self.is_json(real_field):
            return super().to_internal_value(data)
        return data

    def to_representation(self, value):
        real_field = self.get_real_field(value)
        value = super().get_attribute(value)
        if self.is_json(real_field):
            with raise_context():
                value = json.loads(value)
        if real_field:
            return real_field.to_representation(value)
        return value


class DependEnumField(DynamicJsonTypeField):
    """
    Field extends :class:`DynamicJsonTypeField` but its value is not transformed to json and would be given as is.
    Useful for :class:`property` in models or for actions.

    :param field: field in model which value change will change type of current value.
    :type field: str
    :param types: key-value mapping where key is value of subscribed field and
                  value is type (in OpenApi format) of current field.
    :type type: dict
    :param choices: variants of choices for different subscribed field values.
                    Uses mapping where key is value of subscribed field and
                    value is list with values to choice.
    :type choices: dict


    .. note::
        Effective only in GUI. In API works similar to :class:`.VSTCharField` without value modification.
    """
    __slots__ = ()

    def is_json(self, real_field):
        return False


class DependFromFkField(DynamicJsonTypeField):
    """
    Field extends :class:`DynamicJsonTypeField`. Validates field data by :attr:`.field_attribute`
    chosen in related model. By default, any value of :attr:`.field_attribute` validates as :class:`.VSTCharField`.
    To override this behavior set dict attribute ``{field_attribute value}_fields_mapping`` in related model where:

    - **key** - string representation of value type which is received from related instance :attr:`.field_attribute`.
    - **value** - :class:`rest_framework.Field` instance for validation.

    :param field: field in model which value change changes type of current value.
                  Field must be :class:`.FkModelField`.
    :type field: str
    :param field_attribute: attribute of model related model instance with name of type.
    :type field_attribute: str

    .. warning::
        ``field_attribute`` in related model must be :class:`rest_framework.ChoicesField` or
        GUI will show field as simple text.

    """
    __slots__ = ('field', 'field_attribute')

    default_related_field = VSTCharField(allow_null=True, allow_blank=True, default='')

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field')
        self.field_attribute = kwargs.pop('field_attribute')
        super(DynamicJsonTypeField, self).__init__(**kwargs)  # pylint: disable=bad-super-call

    def get_value(self, dictionary: _t.Any) -> _t.Any:
        value = super().get_value(dictionary)
        if value == empty:
            return value  # nocv

        parent_field = self.parent.fields[self.field]  # type: ignore
        related_object: models.Model = parent_field.get_value(dictionary)  # type: ignore
        if not isinstance(related_object, models.Model) and isinstance(parent_field, FkModelField):  # nocv
            # TODO: write test to it
            model_class = get_if_lazy(parent_field.model_class)
            related_object = model_class(**{parent_field.autocomplete_property: related_object})
        related_type = getattr(related_object, self.field_attribute)
        related_field: Field = getattr(
            related_object,
            f'{self.field_attribute}_fields_mapping',
            {related_type: self.default_related_field}
        ).get(related_type, self.default_related_field)
        related_field: Field = copy.deepcopy(related_field)
        related_field.field_name: _t.Text = self.field_name  # type: ignore

        errors: _t.Dict = {}
        primitive_value = related_field.get_value(dictionary)
        try:
            value = related_field.run_validation(primitive_value)
        except ValidationError as exc:
            errors[related_field.field_name] = exc.detail  # type: ignore
        except DjangoValidationError as exc:  # nocv
            errors[related_field.field_name] = get_error_detail(exc)  # type: ignore
        except SkipField:  # nocv
            pass

        if errors:
            raise ValidationError(errors)

        return value

    def get_real_field(self, data) -> _t.Optional[Field]:
        return None


class TextareaField(VSTCharField):
    """
    Field containing multiline string.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = ()


class HtmlField(VSTCharField):
    """
    Field contains html text and marked as format:html. The field does not validate whether its content is HTML.

    .. warning::
        To avoid vulnerability, do not allow users to modify this data because users ate able to execute their scripts.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = ()


class FkField(IntegerField):
    """
    Implementation of ForeignKeyField.You can specify which field of a related model will be
    stored in field(default: "id"), and which will represent field on frontend.

    :param select: OpenApi schema definition name.
    :type select: str
    :param autocomplete_property: this argument indicates which attribute will be
                                  get from OpenApi schema definition model as value.
                                  Default is ``id``.
    :type autocomplete_property: str
    :param autocomplete_represent: this argument indicates which attribute will be
                                   get from OpenApi schema definition model as represent value.
                                   Default is ``name``.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``True``.
    :type use_prefetch: bool
    :param make_link: show value as link to model. Default is ``True``.
    :type make_link: bool
    :param dependence: dictionary, where keys are names of a field from the same model,
                       and keys are name of query filter.
                       If at least one of the fields that we depend on is non nullable, required and set to null,
                       autocompletion list will be empty and field will be disabled.

                       There are some special keys for dependence dictionary to get data that is stored
                       on frontend without additional database query:

                       ``'<<pk>>'``, - get primary key of current instance

                       ``'<<view_name>>'`` get view name from Vue component,

                       ``'<<parent_view_name>>'`` get view name from Vue component,

                       ``'<<view_level>>'`` get view level,

                       ``'<<operation_id>>'`` get operation_id,

                       ``'<<parent_operation_id'>>`` get parent_operation_id
    :type dependence: dict

    Examples:
        .. sourcecode:: python

            field = FkField(select=Category, dependence={'<<pk>>': 'my_filter'})

    This filter  will get pk of current object and make query on frontend '/category?my_filter=3'
    where '3' is primary key of current instance.


    :param filters: dictionary, where keys are names of a field from a related (by this FkField) model,
                    and values are values of that field.
    :type filters: dict

    .. note::
        Intersection of `dependence.values()` and `filters.keys()` will throw error to prevent ambiguous filtering.
    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.
    """
    __slots__ = (
        'select_model',
        'autocomplete_property',
        'autocomplete_represent',
        'use_prefetch',
        'make_link',
        'dependence'
    )

    select_model: _t.Text
    autocomplete_property: _t.Text
    autocomplete_represent: _t.Text
    use_prefetch: bool
    make_link: bool
    dependence: DependenceType
    default_error_messages = {
        'ambiguous_filter': 'Ambiguous filtering, use different filters for dependencies and filters',
    }

    def __init__(self, **kwargs):
        self.select_model = kwargs.pop('select')
        self.autocomplete_property = kwargs.pop('autocomplete_property', 'id')
        self.autocomplete_represent = kwargs.pop('autocomplete_represent', 'name')
        self.use_prefetch = kwargs.pop('use_prefetch', True)
        self.make_link = kwargs.pop('make_link', True)
        self.dependence = kwargs.pop('dependence', None)
        self.filters = kwargs.pop('filters', None)
        super().__init__(**kwargs)
        if self.filters and self.dependence and set(self.dependence.values()) & set(self.filters.keys()):
            self.fail('ambiguous_filter')
        # Remove min/max validators from integer field.
        self.validators = self.validators[:-((self.max_value is not None) + (self.min_value is not None)) or None]


class FkModelField(FkField):
    """
    Extends :class:`.FkField`, but stores referred model class.
    This field is useful for :class:`django.db.models.ForeignKey` fields in model to set.

    :param select: model class (based on :class:`vstutils.models.BModel`) or serializer class
                   which used in API and has path in OpenApi schema.
    :type select: vstutils.models.BModel,vstutils.api.serializers.VSTSerializer
    :param autocomplete_property: this argument indicates which attribute will be
                                  get from OpenApi schema definition model as value.
                                  Default is ``id``.
    :type autocomplete_property: str
    :param autocomplete_represent: this argument indicates which attribute will be
                                   get from OpenApi schema definition model as represent value.
                                   Default is ``name``.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``True``.
    :param make_link: Show value as link to model. Default is ``True``.


    .. warning::
        Model class get object from database during `.to_internal_value` execution. Be careful on mass save executions.

    .. warning::
        Permissions to model which is referred by this field, are not to be checked.
        You should check it manually in signals or validators.

    """

    __slots__ = ('model_class',)

    model_class: _t.Type[models.Model]

    def __init__(self, **kwargs):
        select = kwargs.pop('select')
        if '__extra_metadata__' in dir(select):
            self.model_class = select
            kwargs['select'] = self._get_lazy_select_name_from_model()
        elif isinstance(select, str):
            select = select.split('.')
            assert len(select) == 2, "'select' must match 'app_name.model_name' pattern."
            self.model_class = SimpleLazyObject(lambda: apps.get_model(require_ready=True, *select))
            kwargs['select'] = self._get_lazy_select_name_from_model()
        elif issubclass(select, ModelSerializer):
            self.model_class = select.Meta.model
            kwargs['select'] = getattr(select.Meta, 'ref_name', '') or select.__name__.replace('Serializer', '')
        else:  # nocv
            raise Exception(
                'Argument "select" must be '
                'rest_framework.serializers.ModelSerializer or '
                'vstutils.models.BModel subclass or '
                'string matched "app_name.model_name" pattern.'
            )
        super().__init__(**kwargs)

    def _get_lazy_select_name_from_model(self):
        # pylint: disable=invalid-name
        return SimpleLazyObject(
            lambda: self.model_class.get_list_serializer_name().split('Serializer')[0]
        )

    @functools.lru_cache()
    def _get_data_from_model(self, value):
        self.model_class = get_if_lazy(self.model_class)
        if isinstance(value, self.model_class):
            return value  # nocv
        return self.model_class.objects.get(**{self.autocomplete_property: value})

    def get_value(self, dictionary: _t.Any) -> _t.Any:
        value = super().get_value(dictionary)
        if value is not empty and value is not None:
            value = self._get_data_from_model(value)
        return value

    def to_internal_value(self, data: _t.Union[models.Model, int]) -> _t.Union[models.Model]:  # type: ignore[override]
        if isinstance(data, self.model_class):
            return data
        elif isinstance(data, (int, str)):  # nocv
            # deprecated
            return self._get_data_from_model(data)
        self.fail('Unknown datatype')  # nocv

    def to_representation(self, value: _t.Union[int, models.Model]) -> _t.Any:
        self.model_class = get_if_lazy(self.model_class)
        if self.model_class is not None and isinstance(value, self.model_class._meta.pk.model):  # type: ignore
            return getattr(value, self.autocomplete_property)
        else:  # nocv
            # Uses only if value got from `.values()`
            return value  # type: ignore


class DeepFkField(FkModelField):
    """
    Extends :class:`.FkModelField`, but displays as tree on frontend.

    .. warning::
        This field does not support ``dependence``.
        Use ``filters`` at your own risk, as it would rather break the tree structure.

    :param only_last_child: if True then only allows a value to be selected if it has no children. Default is `False`
    :type only_last_child: bool
    :param parent_field_name: name of parent field in model. Default is `parent`
    :type parent_field_name: str
    """
    def __init__(self, only_last_child: bool = False, parent_field_name='parent', **kwargs):
        super().__init__(**kwargs)
        self.only_last_child = only_last_child
        self.parent_field_name = parent_field_name


class UptimeField(IntegerField):
    """
    Field for some uptime(time duration), in seconds, for example.

    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.

    """

    __slots__ = ()


class RedirectFieldMixin:
    """
    Field mixin indicates that this field is used to send redirect address to frontend after some action.

    :param operation_name: prefix for operation_id, for example if operation_id is `history_get`
           then operation_name is `history`
    :type operation_name: str
    :param depend_field: name of the field that we depend on, its' value will be used for operation_id
    :type depend_field: str
    :param concat_field_name: if True then name of the field will be added at the end of operation_id
    :type concat_field_name: bool
    """
    redirect: bool = True

    __slots__ = ('operation_name', 'depend_field', 'concat_field_name')

    def __init__(self, **kwargs):
        self.operation_name = kwargs.pop('operation_name', None)
        self.depend_field = kwargs.pop('depend_field', None)
        self.concat_field_name = kwargs.pop('concat_field_name', False)
        super().__init__(**kwargs)


class RedirectIntegerField(RedirectFieldMixin, IntegerField):
    """
    Field for redirect by id. Often used in actions for redirect after execution.

    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.

    """

    __slots__ = ()


class RedirectCharField(RedirectFieldMixin, CharField):
    """
    Field for redirect by string. Often used in actions for redirect after execution.

    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.

    """

    __slots__ = ()


class NamedBinaryFileInJsonField(VSTCharField):
    """
    Field that takes JSON with properties:
    * name - string - name of file;
    * mediaType - string - MIME type of file
    * content - base64 string - content of file.

    This field is useful for saving binary files with their names in :class:`django.db.models.CharField`
    or :class:`django.db.models.TextField` model fields. All manipulations with decoding and encoding
    binary content data executes on client. This imposes reasonable limits on file size.

    Additionally, this field can construct :class:`django.core.files.uploadedfile.SimpleUploadedFile`
    from incoming JSON and store it as file in :class:`django.db.models.FileField` if `file` argument is set to `True`

    Attrs:
    :attr:`NamedBinaryInJsonField.file`: if True, accept only subclasses of File as input. If False, accept only string
    input. Default: False.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.

    """

    __slots__ = ('file', 'post_handlers', 'pre_handlers')

    __valid_keys = ('name', 'content', 'mediaType')
    default_error_messages = {
        'not a JSON': lazy_translate('value is not a valid JSON'),  # type: ignore
        'missing key': lazy_translate('key {missing_key} is missing'),  # type: ignore
        'invalid key': lazy_translate('invalid key {invalid_key}'),  # type: ignore
    }
    file_field = FieldFile

    def __init__(self, **kwargs):
        self.file = kwargs.pop('file', False)
        self.post_handlers = kwargs.pop('post_handlers', ())
        self.pre_handlers = kwargs.pop('pre_handlers', ())
        super(NamedBinaryFileInJsonField, self).__init__(**kwargs)

    def validate_value(self, data: _t.Dict):
        if not isinstance(data, dict):
            self.fail('not a JSON')
        invalid_keys = [
            k
            for k in data.keys()
            if k not in self.__valid_keys
        ]

        if invalid_keys:
            self.fail('invalid key', invalid_key=invalid_keys[0])

        for key in self.__valid_keys:
            if key not in data:
                self.fail('missing key', missing_key=key)

    @raise_context_decorator_with_default(default=DEFAULT_NAMED_FILE_DATA)
    def to_representation(self, value) -> _t.Dict[_t.Text, _t.Optional[_t.Any]]:
        if self.file:
            return {'content': value.url, 'name': value.name, 'mediaType': ''}
        else:
            result = json.loads(value)
            if not result.get('mediaType'):
                result['mediaType'] = None
            return result

    def run_validation(self, data=empty):
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data  # nocv
        self.validate_value(data)
        if not self.should_not_handle(data):
            with raise_context(verbose=True):
                data['content'] = base64.b64encode(functools.reduce(
                    lambda binary, func: func(binary, original_data=data),
                    self.pre_handlers,
                    base64.b64decode(data['content'])
                )).decode('utf-8')
        self.run_validators(data)
        return self.to_internal_value(data)

    def run_validators(self, value: _t.Dict) -> None:
        if not raise_context_decorator_with_default(default=False)(self.should_not_handle)(value):
            super().run_validators(value)

    def should_not_handle(self, file):
        return (
            not file['mediaType'] and
            (file['content'].startswith('/') or file['content'].startswith('http')) and
            file['content'].endswith(quote(file['name']))
        )

    def _handle_file(self, file):
        if self.should_not_handle(file) and self.root.instance:
            return self.file_field(
                self.root.instance,
                self.root.instance._meta.get_field(self.field_name),
                file['name']
            )
        content = functools.reduce(
            lambda binary, func: func(binary, original_data=file),
            self.post_handlers,
            base64.b64decode(file['content'])
        )
        return SimpleUploadedFile(
            name=file['name'],
            content=content,
            content_type=file['mediaType']
        )

    def to_internal_value(self, data) -> _t.Any:
        if self.file and data.get('content', None):
            return self._handle_file(data)
        return super().to_internal_value(data)


class NamedBinaryImageInJsonField(NamedBinaryFileInJsonField):
    """
    Extends :class:`.NamedBinaryFileInJsonField` to represent image on frontend
    (if binary image is valid).Validate this field with
    :class:`vstutils.api.validators.ImageValidator`.
    """

    __slots__ = ()


class MultipleNamedBinaryFileInJsonField(NamedBinaryFileInJsonField):
    """
    Extends :class:`.NamedBinaryFileInJsonField` but uses list of JSONs.
    Allows to operate with multiple files as list of :class:`NamedBinaryFileInJsonField`.

    Attrs:
    :attr:`NamedBinaryInJsonField.file`: if True, accept only subclasses of File as input. If False, accept only string
    input. Default: False.
    """

    __slots__ = ()
    default_error_messages = {
        'not a list': lazy_translate('value is not a valid list'),
    }
    file_field = vst_model_fields.MultipleFieldFile

    def run_validators(self, value: _t.Any) -> None:
        if isinstance(value, (list, tuple)):
            for one_value in value:
                super().run_validators(one_value)
        else:  # nocv
            super().run_validators(value)

    def to_internal_value(self, data: _t.List) -> _t.Union[_t.Text, list]:  # type: ignore
        if data is not None:
            if not isinstance(data, list):
                self.fail('not a list')
            for file in data:
                self.validate_value(file)
            if self.file:
                return [
                    self._handle_file(file)
                    for file in data
                ]
        return VSTCharField.to_internal_value(self, data)

    def run_validation(self, data=empty):
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data  # nocv
        self.run_validators(data)
        return self.to_internal_value(data)

    @raise_context_decorator_with_default(default=[])
    def to_representation(self, value) -> _t.List[_t.Dict[_t.Text, _t.Any]]:  # type: ignore
        if self.file:
            return [
                {'content': file.url, 'name': file.name, 'mediaType': ''}
                for file in value
            ]
        return json.loads(value)


class MultipleNamedBinaryImageInJsonField(MultipleNamedBinaryFileInJsonField):
    """
    Extends :class:`.MultipleNamedBinaryFileInJsonField` but uses list of JSONs.
    Used for operating with multiple images and works as list of :class:`NamedBinaryImageInJsonField`.
    """

    __slots__ = ()


class PasswordField(CharField):
    """
    Extends `CharField <https://www.django-rest-framework.org/api-guide/fields/#charfield>`_
    but in schema set format to `password`.
    Show all characters as asterisks instead of real value in GUI.
    """

    def __init__(self, *args, **kwargs):
        kwargs['write_only'] = True
        kwargs['style'] = kwargs.get('style', {})
        kwargs['style']['input_type'] = 'password'
        super(PasswordField, self).__init__(*args, **kwargs)


def _handle_related_value_decorator(func):
    def wrapper(self, value, default, *args, **kwargs):
        if not value:
            return default
        try:
            return func(self, value, default, *args, **kwargs)
        except Exception as err:
            logging.error(err)
            return default
    return wrapper


class RelatedListField(VSTCharField):
    """
    Extends class :class:`.VSTCharField`. With this field you can output reverse ForeignKey relation
    as a list of related instances.

    To use it, you specify 'related_name' kwarg (related_manager for reverse ForeignKey)
    and 'fields' kwarg (list or tuple of fields from related model, which needs to be included).

    By default :class:`.VSTCharField` used to serialize all field values and represent it on
    frontend. You can specify `serializer_class` and override fields as you need. For example title, description
    and other field properties can be set to customize frontend behaviour.

    :param related_name: name of a related manager for reverse foreign key
    :type related_name: str
    :param fields: list of related model fields.
    :type fields: list[str], tuple[str]
    :param view_type: determines how field are represented on frontend. Must be either 'list' or 'table'.
    :type view_type: str
    :param fields_custom_handlers_mapping: includes custom handlers, where key: field_name, value: callable_obj that
                                           takes params: instance[dict], fields_mapping[dict], model, field_name[str]
    :type fields_custom_handlers_mapping: dict
    :param serializer_class: Serializer to customize types of fields, if no serializer provided :class:`.VSTCharField`
                             will be used for every field in `fields` list
    :type serializer_class: type
    """

    def __init__(
            self,
            related_name: _t.Text,
            fields: _t.Union[_t.Tuple, _t.List],
            view_type: str = 'list',
            serializer_class: _t.Optional[_t.Type] = None,
            **kwargs):
        kwargs['read_only'] = True
        kwargs['source'] = "*"
        self.fields_custom_handlers_mapping = kwargs.pop('fields_custom_handlers', {})
        super().__init__(**kwargs)
        # fields for 'values' in qs
        assert isinstance(fields, (tuple, list)), "fields must be list or tuple"
        assert fields, "fields must have one or more values"
        assert view_type in ('list', 'table')
        self._serializer_class = serializer_class
        self.fields = fields
        self.related_name = related_name
        self.view_type = view_type

    @cached_property
    def has_serializer(self):
        return self._serializer_class is not None

    @cached_property
    def serializer_class(self):
        # pylint: disable=import-outside-toplevel
        from .serializers import BaseSerializer

        dependencies = []
        if self.has_serializer:
            dependencies.append(self._serializer_class)
        dependencies.append(type(BaseSerializer)(
            'DefaultRelatedListSerializer',
            (BaseSerializer,),
            {
                f: VSTCharField(allow_null=True, allow_blank=True, default='')
                for f in self.fields
            }
        ))
        return type(dependencies[0])(
            'RelatedListSerializer',
            tuple(dependencies),
            {}
        )

    def get_serializer(self, *args, **kwargs):
        kwargs = {**kwargs}
        kwargs.setdefault('context', self.context)
        return self.serializer_class(*args, **kwargs)

    @_handle_related_value_decorator
    def _handle_named_bin_text(self, value, default):
        return orjson.loads(value)

    @_handle_related_value_decorator
    def _handle_file_or_image_field(self, value, default, field):
        return {
            "content": field.storage.url(value),
            "name": value,
            'mediaType': '',
        }

    @_handle_related_value_decorator
    def _handle_files_or_images_field(self, list_values, default, field):
        return [
            self._handle_file_or_image_field(value, DEFAULT_NAMED_FILE_DATA, field)
            for value in list_values
        ]

    def handle_values_item_iteration(self, instance, fields_mapping, field_name, value, full_field_name=None):
        full_field_name = full_field_name or field_name
        field = fields_mapping.get(field_name, None)

        if '__' in field_name:
            parent_field_name, child_field_name = field_name.split('__', maxsplit=1)
            field = fields_mapping.get(parent_field_name, None)

            if isinstance(field, models.ForeignKey):
                self.handle_values_item_iteration(
                    instance,
                    self.get_model_fields_mapping(field.related_model),
                    child_field_name,
                    value,
                    full_field_name
                )
        elif isinstance(field, vst_model_fields.NamedBinaryFileInJSONField):
            instance[full_field_name] = self._handle_named_bin_text(value, DEFAULT_NAMED_FILE_DATA)

        elif isinstance(field, vst_model_fields.MultipleNamedBinaryFileInJSONField):
            instance[full_field_name] = self._handle_named_bin_text(value, [])

        elif isinstance(field, vst_model_fields.MultipleFileMixin):
            instance[full_field_name] = self._handle_files_or_images_field(value, [], fields_mapping[field_name])

        elif isinstance(field, vst_model_fields.FileField):
            instance[full_field_name] = self._handle_file_or_image_field(
                value,
                DEFAULT_NAMED_FILE_DATA,
                fields_mapping[field_name],
            )

    def handle_values_item(self, instance, fields_mapping, model):
        for field_name, value in instance.items():
            if field_name in self.fields_custom_handlers_mapping:
                self.fields_custom_handlers_mapping[field_name](self, instance, fields_mapping, model, field_name)
            else:
                self.handle_values_item_iteration(instance, fields_mapping, field_name, value)

        return instance

    def get_model_fields_mapping(self, model):
        return {f.name: f for f in model._meta.fields}

    def _prep_data(self, value):
        data = getattr(value, self.related_name).values(*self.fields)
        if self.has_serializer:
            return self.get_serializer(data, many=True).data
        queryset = getattr(value, self.related_name).all()

        handler = functools.partial(
            self.handle_values_item,
            fields_mapping=self.get_model_fields_mapping(queryset.model),
            model=value,
        )
        return tuple(map(handler, data))

    def to_representation(self, value: _t.Type[models.Model]) -> _t.Tuple[_t.Dict]:  # type: ignore[override]
        # get related mapping with id and name of instances
        return lazy(lambda: self._prep_data(value), tuple)()


class RatingField(FloatField):
    """
    Extends class 'rest_framework.serializers.FloatField'. This field represents a rating form input on frontend.
    Grading limits can be specified with 'min_value=' and 'max_value=', defaults are 0 to 5.Minimal step between
    grades are specified in 'step=', default - 1.Frontend visual representation can be chosen
    with 'front_style=', available variants are listed in 'self.valid_front_styles'.

    for 'slider' front style, you can specify slider color, by passing valid color to 'color='.
    for 'fa_icon' front style, you can specify FontAwesome icon that would be used for displaying rating, by passing a
    valid FontAwesome icon code to 'fa_class='.

    :param min_value: minimal level
    :type min_value: float, int
    :param max_value: maximal level
    :type max_value: float, int
    :param step: minimal step between levels
    :type step: float, int
    :param front_style: visualization on frontend field. Allowed: ['stars', 'slider', 'fa_icon'].
    :type front_style: str
    :param color: color of rating element (star, icon or slider) in css format
    :type color: str
    :param fa_class: FontAwesome icon code
    :type fa_class: str
    """
    valid_front_styles = (
        'stars',
        'slider',
        'fa_icon',
    )

    def __init__(
            self,
            min_value: float = 0,
            max_value: float = 5,
            step: float = 1,
            front_style: _t.Text = 'stars',
            **kwargs,
    ):
        assert front_style in self.valid_front_styles, f"front_style should be one of {self.valid_front_styles}"
        self.front_style = front_style
        self.color = kwargs.pop('color', None)
        assert isinstance(self.color, str) or self.color is None, "color should be str"
        self.fa_class = kwargs.pop('fa_class', None)
        assert isinstance(self.fa_class, str) or self.fa_class is None, "fa_class should be str"
        self.step = step
        assert not (step != 1 and front_style != 'slider'), 'custom step can be used only with front_style "slider"'
        super(RatingField, self).__init__(min_value=min_value, max_value=max_value, **kwargs)


def is_all_digits_validator(value):
    if not value.isdigit():
        raise ValidationError(
            translate('This field must contain only digits but input: ') + f'{value}.'
        )


class PhoneField(CharField):
    """
    Extends class 'rest_framework.serializers.CharField'.
    Field for for phone in international format
    """

    def __init__(self, **kwargs):
        kwargs['min_length'] = 8
        kwargs['max_length'] = 16
        super().__init__(**kwargs)

        self.validators.append(is_all_digits_validator)


class MaskedField(CharField):
    """
    Extends class 'rest_framework.serializers.CharField'.
    Field that apply mask to value

    :param mask: `IMask <https://imask.js.org/guide.html>`_
    :type mask: dict, str

    .. note::
        Effective only on frontend.
    """
    def __init__(self, mask, **kwargs):
        super().__init__(**kwargs)
        self.mask = mask


class WYSIWYGField(TextareaField):
    """
    On frontend renders https://ui.toast.com/tui-editor.
    Saves data as markdown and escapes all html tags.
    """

    def to_internal_value(self, data):
        return escape(strip_tags(super().to_internal_value(data)))



import hashlib

from django import template
from django.contrib.auth import get_user_model
from django.templatetags.static import static

User = get_user_model()
register = template.Library()


@register.simple_tag
def get_user_gravatar(user_id):
    """

    Gets link to user's gravatar from serializer.

    Usage::

        {% get_user_gravatar user_id %}

    Examples::

        {% get_user_gravatar 1 %}
        {% get_user_gravatar user.id %}

    """
    try:
        user = User.objects.get(pk=user_id)
    except User.DoesNotExist:
        return static('img/anonymous.png')
    if not user.email:
        return static('img/anonymous.png')
    url_base = 'https://www.gravatar.com/avatar/{}?d=mp'
    user_hash = hashlib.md5(user.email.lower().encode('utf-8')).hexdigest()
    return url_base.format(user_hash)



# pylint: disable=import-error
import collections
import io
import re
import json

import yaml
from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive, directives
from docutils import nodes
from sphinx.util.nodes import nested_parse_with_titles  # type: ignore
from sphinxcontrib.httpdomain import HTTP_STATUS_CODES  # type: ignore


class _YamlOrderedLoader(yaml.SafeLoader):
    pass


_YamlOrderedLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    lambda loader, node: collections.OrderedDict(loader.construct_pairs(node))
)


class VSTOpenApiBase(Directive):
    find_model = re.compile(r"\"\$ref\":.*\"#\/(?P<def_path>.*)\/(?P<def_name>.*)\"")
    find_param = re.compile(r"(?<=\{)[\w]+(?=\})")
    required_arguments = 1  # path to openapi spec
    final_argument_whitespace = True  # path may contain whitespaces
    openapi_version = None
    models_path = 'definitions'
    path_path = 'paths'
    indent_depth = 2
    indent = '   '
    type_dict = {
        'fk': 1,
        'integer': 1,
        'uri': 'http://localhost:8080{}',
        'string': 'example {}',
        'textarea': 'example\ntext\narea\n',
        'boolean': True,
        'select2': 'username',
        'dynamic': 'test_dynamic',
        'uptime': '22:11:34',
        'date_time': '2019-01-07T06:10:31+10:00',
        'html': 'test_html',
        'email': 'example@mail.com',
        'file': 'value data',
        'secretfile': 'secret data',
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__view_list = ViewList()
        self.spec = {}
        self.paths = {}
        self.definitions = {}
        self.current_path = None

    def load_yaml(self):
        """
        Load OpenAPI specification from yaml file. Path to file taking from command
        `vst_openapi`.
        :return:
        """
        env = self.state.document.settings.env
        relpath, abspath = env.relfn2path(directives.path(self.arguments[0]))

        env.note_dependency(relpath)

        encoding = self.options.get('encoding', env.config.source_encoding)
        with io.open(abspath, 'rt', encoding=encoding) as stream:
            spec = yaml.load(stream, _YamlOrderedLoader)  # nosec
        self.spec = spec
        self.paths = spec[self.path_path]
        self.definitions = spec[self.models_path]
        self.openapi_version = spec.get('swagger', None) or spec['openapi']
        self.options.setdefault('uri', 'file://%s' % abspath)

    def write(self, value, indent_depth=0):
        """
        Add lines to ViewList, for further rendering.
        :param value: --line that would be added to render list
        :type value: str, unicode
        :param indent_depth: --value that show indent from left border
        :type indent_depth: integer
        :return:
        """
        self.__view_list.append(self.indent * indent_depth + value, '<openapi>')

    def run(self):
        """
        Main function for prepare and render OpenAPI specification
        :return:
        """
        # Loading yaml
        self.load_yaml()

        # Print paths from schema
        section_title = '**API Paths**'
        self.write(section_title)
        self.write('=' * len(section_title))
        self.print_paths()

        # Print models
        section_title = '**Schemas Description**'
        self.write(section_title)
        self.write('=' * len(section_title))
        self.print_schemas()

        # Render by sphinx
        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, self.__view_list, node)
        return node.children

    def print_paths(self):
        """
        Cycle for prepare information about paths
        :return:
        """
        for path_key, path_value in self.paths.items():
            # Handler for request in path
            self.current_path = path_key
            for request_key, request_value in path_value.items():
                if request_key == 'parameters':
                    continue
                self.get_main_title(path_key, request_key)
                self.get_description(request_value)
                self.get_status_code_and_schema_rst(request_value['responses'])
                self.get_params(path_value['parameters'], 'param')
                self.get_params(request_value['parameters'], 'query')

    def print_schemas(self):
        """
        Print all schemas, one by one
        :return:
        """
        self.indent_depth += 1
        for i in self.definitions:
            def_name = i.split('/')[-1]
            self.write(f'.. _{def_name}:')
            self.write('')
            self.write(f'{def_name} Schema')
            self.write(f'{"`" * (len(def_name) + 7)}')
            self.write('')
            self.write('.. code-block:: json', self.indent_depth)
            self.indent_depth += 1
            self.write('')
            self.definition_rst(def_name)
            self.indent_depth -= 1
            self.write('')
            self.write('')
        self.indent_depth -= 1

    def get_main_title(self, path_name, request_name):
        """
        Get title, from request type and path

        :param path_name: --path for create title
        :type path_name: str, unicode
        :param request_name: --name of request
        :type request_name: str, unicode
        :return:
        """
        main_title = f'.. http:{request_name}:: {path_name}'
        self.write(main_title)
        self.write('')

    def get_status_code_and_schema_rst(self, responses):
        """
        Function for prepare information about responses with example, prepare only
        responses with status code from `101` to `299`
        :param responses: -- dictionary that contains responses, with status code as key
        :type responses: dict
        :return:
        """
        for status_code, response_schema in responses.items():
            status_code = int(status_code)
            schema = response_schema.get('schema', None)
            status = HTTP_STATUS_CODES.get(status_code, None)
            if status is None or not (100 < status_code < 300):
                continue
            self.write('**Example Response**', 1)
            self.write('')
            self.write('.. code-block:: http', 1)
            self.write('')
            self.write(f'HTTP/1.1 {status_code} {status}', 2)
            self.write(f'Vary: {response_schema["description"]}', 2)
            self.write('Content-Type: application/json', 2)
            self.write('')

            if schema:
                self.schema_handler(schema)
            else:
                self.write('{}', self.indent_depth)

    def schema_handler(self, schema):
        """
        Function prepare body of response with examples and create detailed information
        about response fields

        :param schema: --dictionary with information about answer
        :type schema: dict
        :return:
        """
        dict_for_render = schema.get('properties', {}).items()
        if schema.get('$ref', None):
            def_name = schema.get('$ref').split('/')[-1]
            dict_for_render = self.definitions[def_name].get('properties', {}).items()
        elif schema.get('properties', None) is None:
            return ''

        answer_dict = {}
        json_dict = {}
        for opt_name, opt_value in dict_for_render:
            var_type = opt_value.get('format', None) or opt_value.get('type', None) or 'object'
            json_name = self.indent + f':jsonparameter {var_type} {opt_name}:'
            json_dict[json_name] = self.get_json_props_for_response(var_type, opt_value)

            answer_dict[opt_name] = self.get_response_example(opt_name, var_type, opt_value)
            if var_type == 'string':
                answer_dict[opt_name] = answer_dict[opt_name].format(opt_name)

        self.write('')
        for line in json.dumps(answer_dict, indent=4).split('\n'):
            self.write(line, self.indent_depth)

        self.write('')
        for json_param_name, json_param_value in json_dict.items():
            desc = f'{json_param_value["title"]}{json_param_value["props_str"]}' or 'None'
            self.write(json_param_name + ' ' + desc)

    def get_json_props_for_response(self, var_type, option_value):
        """
        Prepare JSON section with detailed information about response

        :param var_type: --contains variable type
        :type var_type: , unicode
        :param option_value: --dictionary that contains information about property
        :type option_value: dict
        :return: dictionary that contains, title and all properties of field
        :rtype: dict
        """
        props = []
        for name, value in option_value.items():
            if var_type in ['dynamic', 'select2']:  # pylint: disable=no-else-break
                break
            elif name in ['format', 'title', 'type']:
                continue
            elif isinstance(value, dict) and value.get('$ref', None):
                props.append(f':ref:`{value["$ref"].split("/")[-1]}`')
            elif '$ref' in name:
                props.append(f':ref:`{value.split("/")[-1]}`')
            elif var_type == 'autocomplete':
                props.append('Example values: ' + ', '.join(value))
            else:
                props.append(f'{name}={value}')

        if len(props):
            props_str = '(' + ', '.join(props) + ')'
        else:
            props_str = ''
        return {'props_str': props_str, 'title': option_value.get('title', '')}

    def get_response_example(self, opt_name, var_type, opt_values):
        """
        Depends on type of variable, return string with example
        :param opt_name: --option name
        :type opt_name: str,unicode
        :param var_type: --type of variable
        :type var_type: str, unicode
        :param opt_values: --dictionary with properties of this variable
        :type opt_values: dict
        :return: example for `var_type` variable
        :rtype: str, unicode
        """
        if opt_name == 'previous' and var_type == 'uri':
            result = None
        elif var_type == 'uri':
            params = {i.group(0): 1 for i in self.find_param.finditer(self.current_path)}
            result = self.type_dict[var_type].format(self.current_path.format(**params))
            if opt_name == 'next':
                result += '?limit=1&offset=1'
        elif opt_name == 'count' and var_type == 'integer':
            result = 2
        elif var_type == 'array':
            items = opt_values.get('items', {}).get('$ref', None)
            item = 'array_example'
            if items:
                item = self.get_object_example(items.split('/')[-1])
            result = [item]
        elif var_type == 'autocomplete':
            result = opt_values.get('enum', [])[0]
        elif var_type in [None, 'object']:
            def_name = (opt_values.get('$ref') or '').split('/')[-1]
            result = self.get_object_example(def_name)
        elif var_type == 'select2':
            def_name = opt_values['x-options']['model']['$ref'].split('/')[-1]
            value_field_name = opt_values['x-options']['value_field']
            def_model = self.definitions[def_name].get('properties')
            value_field = def_model.get(value_field_name, None)
            var_type = value_field.get('format', None) or value_field.get('type', None)
            result = self.get_response_example(opt_name, var_type, def_model)
        else:
            var_type = var_type.replace('-', '_')
            result = opt_values.get('default', None) or self.type_dict.get(var_type, None)
        return result

    def get_object_example(self, def_name):
        """
        Create example for response, from object structure

        :param def_name: --deffinition name of structure
        :type def_name: str, unicode
        :return: example of object
        :rtype: dict
        """
        def_model = self.definitions.get(def_name, {})
        example = {}
        for opt_name, opt_value in def_model.get('properties', {}).items():
            var_type = opt_value.get('format', None) or opt_value.get('type', None)
            example[opt_name] = self.get_response_example(opt_name, var_type, opt_value)
            if var_type == 'string':
                example[opt_name] = example[opt_name].format(opt_name)
        return example

    def definition_rst(self, definition, spec_path=None):
        """
        Prepare and write information about definition
        :param definition: --name of definition that would be prepared for render
        :type definition: str, unicode
        :param spec_path: --path to definitions
        :type spec_path: str, unicode
        :return:
        """
        spec_path = spec_path or self.models_path
        definitions = self.spec[spec_path]
        definition_property = definitions[definition]['properties'].copy()
        if not definition_property:
            self.write('{}', self.indent_depth)
            return
        self.indent_depth += 1
        definition_property = self.find_nested_models(definition_property, definitions)
        json_str = json.dumps(definition_property, indent=4)
        for line in json_str.split('\n'):
            self.write(line, self.indent_depth)
        self.indent_depth -= 1

    def find_nested_models(self, model, definitions):
        """
        Prepare dictionary with reference to another definitions, create one dictionary
        that contains full information about model, with all nested reference
        :param model: --dictionary that contains information about model
        :type model: dict
        :param definitions: --dictionary that contains copy of all definitions
        :type definitions: dict
        :return: dictionary with all nested reference
        :rtype: dict
        """
        for key, value in model.items():
            if isinstance(value, dict):
                model[key] = self.find_nested_models(value, definitions)
            elif key == '$ref':
                def_name = value.split('/')[-1]
                def_property = definitions[def_name]['properties']
                return self.find_nested_models(def_property, definitions)
        return model

    def get_params(self, params, name_request):
        """
        Prepare and add for further render parameters.
        :param params: --dictionary with parameters
        :type params: dict
        :param name_request: --type of the parameters
        :type name_request: str, unicode
        :return:
        """
        self.write('')
        for elem in params:
            request_type = elem['type'] if elem.get('type', None) else 'schema'
            name = elem['name']
            if elem.get('required', None):
                name += '(required)'
            schema = elem.get('schema', None)
            name = f':{name_request} {request_type} {name}:'
            if schema:
                definition = schema['$ref'].split('/')[-1]
                self.write(name + f' :ref:`{definition}`', 1)
                self.write('')
            else:
                desc = elem.get('description', '')
                self.write(name)
                self.write(f'{desc}', self.indent_depth + 1)
        self.write('')

    def get_description(self, request_value):
        """
        Add description about this path and type
        :param request_value: --dictionary with information about this path
        :type request_value: dict
        :return:
        """
        self.write(request_value['description'], 1)
        self.write('')


def setup(app):
    app.setup_extension('sphinxcontrib.httpdomain')
    app.add_directive('vst_openapi', VSTOpenApiBase)



from typing import List, Dict
from . import base


class BulkViewSet(base.rvs.APIView):
    client_environ_keys_copy: List
    type_to_bulk: Dict


class HealthView(base.ListNonModelViewSet):
    ...



from typing import Dict

from django.urls import reverse
from django.conf import settings
from django.http.request import HttpRequest
from django.utils.functional import lazy, SimpleLazyObject

from ..utils import import_class, raise_context_decorator_with_default
from ..static_files import SPA_STATIC_FILES_PROVIDERS


manifest_object = import_class(settings.MANIFEST_CLASS)()
project_lib_version = getattr(settings, 'PROJECT_LIB_VERSION', '')
project_version = getattr(settings, 'PROJECT_VERSION', '')
vstutils_version = settings.VSTUTILS_VERSION
gui_version = "_".join(map(str, [project_version, project_lib_version, vstutils_version]))
static_path = getattr(settings, 'STATIC_URL', '/static/')
static_list = SPA_STATIC_FILES_PROVIDERS.get_sorted_list()
debug_enabled = getattr(settings, 'DEBUG', False)
ignore_errors_decorator = raise_context_decorator_with_default(default={})


def lazy_decorator(func):
    return lazy(ignore_errors_decorator(func), dict)


def lazy_value(func):
    return SimpleLazyObject(func)


def static_file_set_version_to_name(files_list_object: dict):
    new_object = files_list_object.copy()
    version_source = new_object.get('source', 'gui') or 'gui'
    new_object['version'] = globals().get(version_source + '_version', gui_version)
    return new_object


@lazy_decorator
def settings_constants(request: HttpRequest) -> Dict:
    # pylint: disable=unused-argument
    return {
        "login_url": reverse('login'),
        "logout_url": reverse('logout'),
        "docs_url": reverse('docs:docs_root') if getattr(settings, 'HAS_DOCS', False) else '',
        "has_docs": getattr(settings, 'HAS_DOCS', False),
        "timezone": getattr(settings, 'TIME_ZONE', 'UTC'),
        "debug": debug_enabled,
        "languages": settings.LANGUAGES,
    }


@lazy_decorator
def project_args(request: HttpRequest) -> Dict:
    host_url = request.build_absolute_uri('/')[:-1]
    ver_key = f'{getattr(settings, "VST_PROJECT", "vstutils")}_version'
    request_gui_version = f'{gui_version}_{str(request.user.id or 0)}'
    return {
        "host_url": host_url,
        "gui_version": gui_version,
        "gui_named_version": f'{settings.VST_PROJECT}_{gui_version}',
        "gui_user_version": request_gui_version,
        "vstutils_version": settings.VSTUTILS_VERSION,
        "project_lib_version": project_lib_version,
        "project_version": project_version,
        ver_key: project_version,
        "project_gui_name": getattr(settings, 'PROJECT_GUI_NAME', None),
        "project_menu": getattr(settings, 'PROJECT_GUI_MENU', []),
        "openapi_url": reverse('endpoint'),
        "endpoint_path": reverse('endpoint'),
        "api_version": settings.VST_API_VERSION,
        "api_url": f'{host_url}{reverse("api-root")}{settings.VST_API_VERSION}/',
        "enable_gravatar": settings.ENABLE_GRAVATAR,
        "registration_enabled": settings.REGISTRATION_ENABLED,
    }


@lazy_decorator
def pwa_context(request: HttpRequest) -> Dict:
    return {
        "manifest_object": manifest_object,
        "static_files_list": lazy_value(
            lambda: tuple(map(
                static_file_set_version_to_name,
                static_list
            ))
        ),
        "block_timeout": 86400 if not debug_enabled else 0,
    }


@lazy_decorator
def headers_context(request: HttpRequest) -> Dict:
    result = dict(request.META)
    result['HTTP_X_APP'] = result.get('HTTP_X_APP', 'browser')
    return {'metadata': result}



import sys
from subprocess import check_call

from ._base import DockerCommand


class Command(DockerCommand):
    interactive = True
    help = "Run uwsgi server with configuration from ENV."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            'args',
            metavar='uwsgi_arg=value', nargs='*',
            help='Args "name=value" uwsgi server.',
        )

    def handle(self, *args, **options):
        super().handle(*args, **options)

        if not self.config['docker'].getboolean('migration', fallback=True):
            success, error = True, ''  # nocv
        else:
            success, error = self.migrate(options)
        if success:
            try:
                check_call(
                    [sys.executable, '-m', self.project_name, self._settings('WEBSERVER_COMMAND', 'web')],
                    env=self.env, bufsize=0, universal_newlines=True,
                )
            except KeyboardInterrupt:  # nocv
                self._print('Exit by user...', 'WARNING')
                return
        else:
            self._print(error, 'ERROR')
            sys.exit(10)



import typing as _t
import logging
import traceback

from django.core.cache import cache
from django.contrib.auth import get_user_model, backends
from django.db.models import signals, QuerySet
from django.dispatch import receiver
from django.conf import settings
from django.http.request import HttpRequest

from .utils import SecurePickling, ObjectHandlers, raise_context, raise_context_decorator_with_default
try:
    from .ldap_utils import LDAP as _LDAP
    HAS_LDAP = True
except ImportError:  # nocv
    _LDAP = object
    HAS_LDAP = False

UserModel = get_user_model()
AuthRes = _t.Optional[UserModel]
logger = logging.getLogger(settings.VST_PROJECT_LIB)
user_cache_prefix = 'auth_user_id_val'
secure_serializer = SecurePickling()


if settings.CACHE_AUTH_USER:
    @receiver(signals.post_save, sender=UserModel)
    @receiver(signals.post_delete, sender=UserModel)
    def invalidate_user_from_cache(instance: UserModel, created=False, *args, **kwargs):
        if created:
            return
        cache.delete(f'{user_cache_prefix}_{instance.id}')


@raise_context_decorator_with_default(default=None)
def get_secured_cache(key) -> AuthRes:
    user = cache.get(key)
    if isinstance(user, UserModel):
        return user  # nocv
    elif isinstance(user, str):
        return secure_serializer.loads(user)


@raise_context()
def set_secured_cache(key, value):
    cache.set(key, secure_serializer.dumps(value))


def cache_user_decorator(func):
    if not settings.CACHE_AUTH_USER:
        return func  # nocv

    def wrapper(backend, user_id: int):
        cache_key = f'{user_cache_prefix}_{user_id}'
        user = get_secured_cache(cache_key)
        if user is None:
            user = func(backend, user_id)
        if isinstance(user, UserModel):
            set_secured_cache(cache_key, user)
        return user

    return wrapper


class BaseAuthBackend(backends.ModelBackend):
    def authenticate(self, request: HttpRequest, username=None, password=None, **kwargs):
        raise NotImplementedError  # nocv

    def patch_user_queryset(self, queryset: QuerySet) -> QuerySet:
        return queryset

    @cache_user_decorator
    def get_user(self, user_id: int) -> AuthRes:
        # pylint: disable=protected-access
        try:
            user = self.patch_user_queryset(UserModel._default_manager.select_related('twofa')).get(pk=user_id)
        except UserModel.DoesNotExist:  # nocv
            return None
        return user if self.user_can_authenticate(user) else None


class LDAP(_LDAP):
    """
    LDAP class wrapper
    """


class LdapBackend(BaseAuthBackend):  # nocv
    @property
    def domain(self):
        return settings.LDAP_DOMAIN

    @property
    def server(self):
        return settings.LDAP_SERVER

    def authenticate(self, request: HttpRequest, username: _t.Text = None, password: _t.Text = None) -> AuthRes:
        # pylint: disable=protected-access,unused-argument
        if not self.server or not HAS_LDAP:
            return
        try:
            backend = LDAP(self.server, username, password, self.domain)
            if not backend.isAuth():
                return
            user = self.patch_user_queryset(UserModel._default_manager.select_related('twofa'))\
                .get_by_natural_key(backend.domain_user)
            if self.user_can_authenticate(user) and backend.isAuth():
                return user
        except:
            logger.debug(traceback.format_exc())
            return


class AuthPluginsBackend(BaseAuthBackend):
    auth_handlers = ObjectHandlers('AUTH_PLUGINS')
    auth_header = 'HTTP_X_AUTH_PLUGIN'

    @raise_context()
    def auth_with_plugin(self, plugin: _t.Text, request: HttpRequest, username: _t.Text, password: _t.Text) -> AuthRes:
        return self.auth_handlers.get_object(plugin).authenticate(request, username, password)

    @raise_context()
    def authenticate(self, request: HttpRequest, username: _t.Text = None, password: _t.Text = None) -> AuthRes:
        # pylint: disable=protected-access,unused-argument
        if request and self.auth_header in request.META:
            return self.auth_with_plugin(
                request.META[self.auth_header], request, username, password
            )
        for plugin_name in self.auth_handlers.list():
            result = self.auth_with_plugin(plugin_name, request, username, password)
            if result:
                return result



import sys

from ._base import DockerCommand


class Command(DockerCommand):
    interactive = True
    help = "Run migration in docker container."

    def handle(self, *args, **options):
        super().handle(*args, **options)

        success, error = self.migrate(options)
        if not success:
            self._print(error, 'ERROR')
            sys.exit(10)



import typing as _t

import pyotp
from django.contrib.auth import get_user_model, update_session_auth_hash
from django.contrib.auth.hashers import make_password
from django.contrib.auth.password_validation import validate_password
from django.contrib.auth.models import AbstractUser
from django.db import transaction
from django.conf import settings
from django.utils.functional import SimpleLazyObject, cached_property
from django_filters import BooleanFilter, CharFilter
from rest_framework import serializers, exceptions, request as drf_request
from vstutils.api import fields, base, permissions, responses, decorators as deco
from vstutils.api.filters import DefaultIDFilter, name_filter, name_help
from vstutils.api.serializers import VSTSerializer, DataSerializer, BaseSerializer
from vstutils.api.models import TwoFactor, RecoveryCode, UserSettings
from vstutils.utils import raise_context_decorator_with_default, translate, lazy_translate as __

User: _t.Type[AbstractUser] = get_user_model()  # type: ignore[override]


class ChangePasswordPermission(permissions.IsAuthenticatedOpenApiRequest):
    def has_object_permission(self, request: drf_request.Request, view: base.GenericViewSet, obj: User):  # type: ignore
        return (
            request.user.is_superuser or
            (
                isinstance(obj, request.user.__class__) and
                request.user.pk == obj.pk  # type: ignore
            )
        )


class UserSerializer(VSTSerializer):
    is_active = serializers.BooleanField(default=True)
    is_staff = serializers.BooleanField(default=False)
    email = serializers.EmailField(required=False)

    class UserExist(exceptions.ValidationError):
        status_code = 409

    class Meta:
        model = User
        fields = (
            'id',
            'username',
            'is_active',
            'is_staff',
            'email',
        )
        read_only_fields = ('is_superuser',)
        ref_name = 'User'

    @cached_property
    @raise_context_decorator_with_default(default=False)
    def is_staff_or_super(self):
        return self.context['request'].user.is_staff or self.context['request'].user.is_superuser

    def create(self, validated_data):
        """ Create user from validated data. """

        if not self.is_staff_or_super:
            raise exceptions.PermissionDenied  # nocv

        validated_data['password'] = make_password(validated_data['password'])
        return super().create(validated_data)

    def update(self, instance, validated_data):
        if not self.is_staff_or_super and instance.id != self.context['request'].user.id:
            # can't be tested because PATCH from non privileged user to other
            # user fails at self.get_object() in View
            raise exceptions.PermissionDenied  # nocv
        instance.username = validated_data.get('username', instance.username)
        instance.is_active = validated_data.get('is_active', instance.is_active)
        instance.email = validated_data.get('email', instance.email)
        instance.first_name = validated_data.get('first_name', instance.first_name)
        instance.last_name = validated_data.get('last_name', instance.last_name)
        instance.is_staff = validated_data.get('is_staff', instance.is_staff)
        instance.save()
        return instance


class OneUserSerializer(UserSerializer):
    class Meta:
        model = User
        fields: _t.Union[_t.List[_t.Text], _t.Tuple] = (
            'id',
            'username',
            'is_active',
            'is_staff',
            'first_name',
            'last_name',
            'email',
        )
        read_only_fields = (
            'is_superuser',
            'date_joined',
        )


class CreateUserSerializer(OneUserSerializer):
    password = fields.VSTCharField(write_only=True)
    password2 = fields.VSTCharField(write_only=True, label='Repeat password')

    class Meta(OneUserSerializer.Meta):
        fields = list(OneUserSerializer.Meta.fields) + ['password', 'password2']

    def run_validation(self, data=serializers.empty):
        validated_data = super().run_validation(data)
        if validated_data['password'] != validated_data.pop('password2', None):
            raise exceptions.ValidationError({
                "password": [translate('Passwords do not match.')],
            })
        return validated_data


class ChangePasswordSerializer(DataSerializer):
    old_password = fields.PasswordField(required=True)
    password = fields.PasswordField(required=True, label='New password')
    password2 = fields.PasswordField(required=True, label='Confirm new password')

    def update(self, instance, validated_data):
        if not instance.check_password(validated_data['old_password']):
            raise exceptions.AuthenticationFailed()
        if validated_data['password'] != validated_data['password2']:
            raise exceptions.ValidationError(
                translate("New passwords values are not equal.")
            )
        validate_password(validated_data['password'])
        instance.set_password(validated_data['password'])
        instance.save()
        return instance

    def to_representation(self, instance):
        return {
            'old_password': '***',
            'password': '***',
            'password2': '***'
        }


class TwoFASerializer(VSTSerializer):
    enabled = serializers.BooleanField(read_only=True)
    secret = serializers.CharField(write_only=True, default=None, allow_null=True)
    pin = serializers.CharField(write_only=True, required=False, label='Enter the six-digit code from the application')
    recovery = serializers.CharField(write_only=True, required=False)

    default_error_messages = {
        'invalid_pin': __('Invalid authentication code'),  # type: ignore
        'no_secret_provided': __('Secret string must be provided'),  # type: ignore
    }

    class Meta:
        model = TwoFactor
        fields = (
            'enabled',
            'secret',
            'pin',
            'recovery',
        )

    def verify_pin(self, secret, pin):
        if not secret:
            self.fail('no_secret_provided')
        if not pyotp.TOTP(secret).verify(pin):
            self.fail('invalid_pin')

    @transaction.atomic()
    def create(self, validated_data: dict):
        recovery_codes = validated_data.pop('recovery', '')
        self.verify_pin(validated_data.get('secret'), validated_data.pop('pin', ''))
        instance = super().create(validated_data)
        RecoveryCode.objects.bulk_create([
            RecoveryCode(key=code, tfa=instance)
            for code in tuple(filter(bool, recovery_codes.split(',')))[:15]
        ])
        return instance

    def update(self, instance, validated_data):
        instance.delete()
        instance.secret = None  # type: ignore
        return instance


class LanguageFromRequestDefault:
    requires_context = True

    def __call__(self, serializer_field):
        return serializer_field.context['request'].language.code


class MainSettingsSerializer(BaseSerializer):
    language = serializers.ChoiceField(choices=settings.LANGUAGES, default=LanguageFromRequestDefault())  # type: ignore
    dark_mode = serializers.BooleanField(default=False)

    class Meta:
        ref_name = '_MainSettings'


class UserSettingsSerializer(BaseSerializer):
    main = MainSettingsSerializer(default={})
    custom = serializers.JSONField(default={})

    class Meta:
        ref_name = '_UserSettings'


class UserFilter(DefaultIDFilter):
    is_active = BooleanFilter(help_text=__('Boolean value meaning status of user.'))
    first_name = CharFilter(help_text=__('Users first name.'))
    last_name = CharFilter(help_text=__('Users last name.'))
    email = CharFilter(help_text=__("Users e-mail value."))
    username__not = CharFilter(method=name_filter, help_text=name_help, label=__('Username not'))
    username = CharFilter(method=name_filter, help_text=name_help)

    class Meta:
        model = User
        fields = (
            'id',
            'username',
            'is_active',
            'first_name',
            'last_name',
            'email',
        )


class UserViewSet(base.ModelViewSet):
    """
    API endpoint that allows users to be viewed or edited.
    """
    # pylint: disable=invalid-name

    model: _t.Type[AbstractUser] = User
    serializer_class: _t.Type[UserSerializer] = UserSerializer
    serializer_class_one: _t.Type[OneUserSerializer] = OneUserSerializer
    serializer_class_create: _t.Type[CreateUserSerializer] = CreateUserSerializer
    serializer_class_change_password: _t.Type[DataSerializer] = ChangePasswordSerializer
    serializer_class_twofa: _t.Type[serializers.BaseSerializer] = TwoFASerializer
    serializer_class__settings: _t.Type[serializers.BaseSerializer] = UserSettingsSerializer
    filterset_class = UserFilter
    permission_classes = (permissions.SuperUserPermission,)
    optimize_get_by_values = False

    @property
    def search_fields(self):
        return tuple(
            filter(
                self.filterset_class.Meta.fields.__contains__,
                ('username', 'email', 'phone', 'first_name', 'last_name')
            )
        )

    def get_object(self) -> AbstractUser:
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        if self.kwargs.get(lookup_url_kwarg, None) == 'profile':
            self.kwargs[lookup_url_kwarg] = getattr(self.request.user, self.lookup_field)
        return super().get_object()

    def destroy(self, request: drf_request.Request, *args, **kwargs):
        user = self.get_object()
        if user == request.user:
            return responses.HTTP_409_CONFLICT("Could not remove youself.")
        return super().destroy(request, *args, **kwargs)  # pylint: disable=no-member

    @transaction.atomic
    def partial_update(self, request: drf_request.Request, *args, **kwargs):
        return self.update(request, partial=True)

    @transaction.atomic
    def update(self, request: drf_request.Request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)
        return responses.HTTP_200_OK(serializer.data)

    @deco.action(["post"], detail=True, permission_classes=(ChangePasswordPermission,))
    def change_password(self, request: drf_request.Request, *args, **kwargs):
        user = self.get_object()
        serializer = self.get_serializer(user, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        update_session_auth_hash(request, user)
        return responses.HTTP_201_CREATED(serializer.data)

    @deco.action(['get', 'put'], detail=True, permission_classes=(ChangePasswordPermission,))
    def twofa(self, request: drf_request.Request, *args, **kwargs):
        user: User = self.get_object()  # type: ignore
        instance = getattr(user, 'twofa', None)

        if request.method.upper() == 'PUT':  # type: ignore
            serializer = self.get_serializer(instance, data=request.data)
            serializer.is_valid(raise_exception=True)
            serializer.save(user=user)
            request.session['2fa'] = True
        else:
            serializer = self.get_serializer(
                instance or
                SimpleLazyObject(lambda: TwoFactor(user=user))
            )

        return responses.HTTP_200_OK(serializer.data)

    @deco.action(['get', 'put'], detail=True, permission_classes=(ChangePasswordPermission,))
    def _settings(self, request: drf_request.Request, *args, **kwargs):
        instance, _ = UserSettings.objects.get_or_create(
            user=self.get_object(),
            defaults=SimpleLazyObject(lambda: {'value': self.get_serializer({}).data})  # type: ignore
        )

        if request.method.upper() == 'PUT':  # type: ignore
            serializer = self.get_serializer(instance.value, data=request.data)
            serializer.is_valid(raise_exception=True)
            instance.value = serializer.save()
            instance.save()
        else:
            serializer = self.get_serializer(instance.value)

        return responses.HTTP_200_OK(serializer.data)



from rest_framework.metadata import SimpleMetadata

from . import fields, serializers


class VSTMetadata(SimpleMetadata):
    label_lookup = SimpleMetadata.label_lookup
    origin_mapping = SimpleMetadata.label_lookup.mapping
    mapping_fields = {
        fields.FileInStringField: 'file',
        fields.SecretFileInString: 'secretfile',
        fields.AutoCompletionField: 'autocomplete',
        serializers.JsonObjectSerializer: 'json',
        fields.DependEnumField: 'dynamic',
        fields.HtmlField: 'html',
    }
    label_lookup.mapping.update(mapping_fields)



# pylint: disable=unused-import
import typing as _t
import random  # noqa: F401
import string  # noqa: F401
import os  # noqa: F401
import uuid
from unittest.mock import patch, Mock
import json  # noqa: F401

import ormsgpack
from django.apps import apps
from django.db import transaction
from django.urls import reverse
from django.test import TestCase, override_settings  # noqa: F401
from django.contrib.auth import get_user_model

from .utils import import_class, raise_context_decorator_with_default

User = get_user_model()

BulkDataType = _t.Union[_t.List[_t.Dict[_t.Text, _t.Any]], str, bytes, bytearray]
ApiResultType = _t.Union[BulkDataType, _t.Dict, _t.Sequence[BulkDataType]]


class BaseTestCase(TestCase):
    """
    Main testcase class extends :class:`django.test.TestCase`.
    """
    server_name = 'vstutilstestserver'

    #: Attribute with default project models module.
    models = None

    #: Default http status codes for different http methods. Uses in :meth:`.get_result`
    std_codes: _t.Dict[_t.Text, int] = {
        'get': 200,
        'post': 201,
        'patch': 200,
        'delete': 204
    }
    use_msgpack = False

    class user_as:
        # pylint: disable=invalid-name
        """
        Context for execute bulk or something as user.
        The context manager overrides ``self.user`` in TestCase and revert this
        changes on exit.

        :param user: new user object for execution.
        :type user: django.contrib.auth.models.AbstractUser
        """

        def __init__(self, testcase, user):
            self.testcase = testcase
            self.user = user

        def __enter__(self):
            self.old_user = self.testcase.user
            self.testcase.user = self.user
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.testcase.user = self.old_user

    def _create_user(self, is_super_user=True, **kwargs):
        username = kwargs.pop('username', self.random_name())
        email = kwargs.pop('email', username + '@gmail.com')
        password = kwargs.pop('password', username.upper())
        if is_super_user:
            user = User.objects.create_superuser(username=username,
                                                 password=password,
                                                 email=email,
                                                 **kwargs)
        else:
            user = User.objects.create_user(username=username,
                                            password=password,
                                            email=email,
                                            **kwargs)
        user.data = {'username': username, 'password': password}
        return user

    def _login(self):
        client = self.client
        client.force_login(self.user)
        return client

    def _logout(self, client):
        self.assertEqual(client.get(self.logout_url).status_code, 302)

    def _check_update(self, url, data, **fields):
        """
        Test update instance of model

        :param url: url to instance
        :param data: update fields
        :param fields: checking resulted fields as named args
        :return: None

        """
        self.get_result(fields.pop('method', 'patch'), url, data=json.dumps(data))
        result = self.get_result("get", url)
        self.assertTrue(isinstance(result, dict))
        for field, value in fields.items():
            self.assertEqual(result[field], value)

    def __get_rendered(self, response):
        # pylint: disable=protected-access
        try:
            media_type = f'{getattr(response, "accepted_media_type", "")}' or \
                         response._content_type_for_repr.split(";")[0].replace('"', '').replace(',', '').strip()
            rendered_content = (
                getattr(response, "rendered_content", False) or response.content
            )
            if media_type == 'application/msgpack':
                return ormsgpack.unpackb(rendered_content)
            if getattr(rendered_content, 'decode', False):
                rendered_content = str(rendered_content.decode('utf-8'))
            try:
                if media_type in ('application/json', 'application/openapi+json'):
                    return json.loads(rendered_content)
                raise Exception
            except:
                return str(rendered_content)
        except ValueError:  # nocv
            return None

    def setUp(self):
        # pylint: disable=import-outside-toplevel
        from django.conf import settings
        self.settings_obj = settings
        self.client_kwargs = {
            "HTTP_X_FORWARDED_PROTOCOL": 'https',
            'SERVER_NAME': self.server_name
        }
        self.client = self.client_class(**self.client_kwargs)
        self.user = self._create_user()
        self.login_url = reverse('login')
        self.confirmation_redirect_url = self.login_url + '?confirmation=true'
        self.logout_url = reverse('logout')
        self.last_response = None

    def _settings(self, item, default=None):
        return getattr(self.settings_obj, item, default)

    def random_name(self) -> _t.Text:
        """
        Simple function which returns uuid1 string.
        """
        return str(uuid.uuid1())

    def call_registration(self, data, **kwargs):
        """
        Function for calling registration. Just got form data and headers.

        :param data: Registration form data.
        :type data: dict
        :param kwargs: named arguments with request headers.
        """

        client = self.client_class(**self.client_kwargs)
        return client.post(reverse('user_registration'), data=data, **kwargs)

    def get_url(self, *items) -> _t.Text:
        """
        Function for creating url path based on ``VST_API_URL`` and ``VST_API_VERSION`` settings.
        Without arguments returns path to default version of api.

        :return: string like ``/api/v1/.../.../`` where ``...`` is args of function.
        """

        vst_api_url = self._settings("VST_API_URL")
        vst_api_version = self._settings("VST_API_VERSION")

        items = items[2:] if items[:2] == (vst_api_url, vst_api_version) else items

        url = f'/{vst_api_url}/{vst_api_version}/'
        url += "".join(f'{i}/' for i in filter(bool, items))

        return url

    @classmethod
    def patch(cls, *args, **kwargs) -> _t.ContextManager[Mock]:
        """
        Simple :func:`unittest.mock.patch` class-method wrapper.
        """
        return patch(*args, **kwargs)  # type: ignore

    def get_model_class(self, model):
        """
        Getting model class by string or return model arg.

        :param model: string which contains model name (if attribute ``model`` is set to the testcase class),
                      module import, ``app.ModelName`` or :class:`django.db.models.Model`.
        :type model: str,django.db.models.Model
        :return: Model class.
        :rtype: django.db.models.Model

        """

        handlers = (
            lambda x: getattr(self.models, x, None) if self.models is not None else None,
            import_class,
            apps.get_model
        )

        if isinstance(model, str):
            for handler in map(raise_context_decorator_with_default(default=None), handlers):
                result = handler(model)
                if result:
                    model = result
                    break

        assert not isinstance(model, str), f"couldn't find '{model}'."
        return model

    def get_model_filter(self, model, **kwargs):
        """
        Simple wrapper over :meth:`.get_model_class` which returns filtered
        queryset from model.

        :param model: string which contains model name (if attribute ``model`` is set to the testcase class),
                      module import, ``app.ModelName`` or :class:`django.db.models.Model`.
        :type model: str,django.db.models.Model
        :param kwargs: named arguments to :meth:`django.db.models.query.QuerySet.filter`.
        :rtype: django.db.models.query.QuerySet

        """
        return self.get_model_class(model).objects.filter(**kwargs)

    def get_count(self, model, **kwargs):
        """
        Simple wrapper over :meth:`.get_model_filter` which returns counter of items.

        :param model: string which contains model name (if attribute ``model`` is set to the testcase class),
                      module import, ``app.ModelName`` or :class:`django.db.models.Model`.
        :type model: str,django.db.models.Model
        :param kwargs: named arguments to :meth:`django.db.models.query.QuerySet.filter`.
        :return: number of instances in database.
        :rtype: int
        """
        return self.get_model_filter(model, **kwargs).count()

    def change_identity(self, is_super_user=False):
        old_user = self.user
        self.user = self._create_user(is_super_user)
        return old_user

    def render_api_response(self, response):
        return self.__get_rendered(response)

    @transaction.atomic
    def result(self, request, url, code=200, *args, **kwargs):
        response = request(url, secure=True, *args, **kwargs)
        self.last_response = response
        self.assertRCode(response, code, url)
        return self.render_api_response(response)

    def assertCount(self, iterable: _t.Sized, count: int, msg: _t.Any = None):
        """
        Call :func:`len` over ``iterable`` and check equals with ``count``.

        :param iterable: any iterable object which could be sended to :func:`len`.
        :param count: expected result.
        :param msg: error message

        """
        self.assertEqual(len(iterable), count, msg)

    def assertRCode(self, resp, code=200, *additional_info):
        """
        Fail if response code is not equal. Message is response body.

        :param resp: response object
        :type resp: django.http.HttpResponse
        :param code: expected code
        :type code: int
        """
        err_msg = "{} != {}\n{}\n{}".format(
            resp.status_code, code,
            self.__get_rendered(resp),
            self.user
        )
        if additional_info:
            err_msg += '\nAdditional info:\n'
            err_msg += '\n'.join([str(i) for i in additional_info])
        self.assertEqual(resp.status_code, code, err_msg)

    def assertCheckDict(self, first: _t.Dict, second: _t.Dict, msg: _t.Text = None):
        """
        Fail if the two fields in dicts are unequal as determined by the '==' operator.
        Checks if first not contains or not equal field in second
        """
        for field_name in first.keys():
            self.assertEqual(
                first[field_name], second.get(field_name, None), msg or [first, second]
            )

    def post_result(self, url, code=None, *args, **kwargs):
        return self.get_result("post", url, code, *args, **kwargs)

    def get_result(self, rtype, url, code: int = None, *args, **kwargs) -> ApiResultType:
        """
        Execute and test response code on request with returning parsed result of request.
        The method uses the following procedure:

        - Test client authorization (with :attr:`.user` which creates in :meth:`.setUp`).
        - Executing a request (sending args and kwargs to request method).
        - Parsing the result (converts json string to python-object).
        - Checking the http status code with :meth:`.assertRCode`
          (if you have not specified it,
          the code will be selected in accordance with the request method
          from the standard set :attr:`.std_codes`).
        - Logout client.
        - Return parsed result.

        :param rtype:     request type (methods from Client cls): get, post etc.
        :param url:       requested url string or tuple for :meth:`.get_url`.
                          You can use :meth:`.get_url` for url building or setup it as full string.
        :param code:      expected return code from request.
        :param relogin:   execute force login and logout on each call. Default is ``True``.
        :param args:      extra-args for Client class request method.
        :param kwargs:    extra-kwargs for Client class request method.
        :return:          result of request.

        """

        relogin = kwargs.get('relogin', True)

        if relogin:
            client = self._login()
        else:
            client = self.client

        request_handler = getattr(client, rtype)

        if kwargs.get("data", False):
            if isinstance(kwargs["data"], str):
                kwargs["content_type"] = "application/json"
            elif isinstance(kwargs["data"], (bytes, bytearray)):
                kwargs["content_type"] = "application/msgpack"

        if 'content_type' in kwargs and kwargs["content_type"].startswith('application/'):
            kwargs['HTTP_ACCEPT'] = kwargs["content_type"]

        kwargs['code'] = code or self.std_codes.get(rtype, 200)

        if isinstance(url, (tuple, list)):
            url = self.get_url(*url)

        result = self.result(request_handler, url, *args, **kwargs)

        if relogin:
            self._logout(client)

        return result

    def mass_create(self, url, data, *fields, **kwargs):
        """
        Mass creation objects in api-abstration. Uses :meth:`.get_result` method.

        :param url: url to abstract layer like argument in :meth:`.get_result`. For example: ``/api/v1/project/``.
        :param data: list with data to send on creation and feature checks.
        :params fields: list of fields to check after creation.
        :param kwargs: extra-kwargs for request method.
        :return: list of id from all resulted objects.
        :rtype: list

        .. note::
            The method does not use endpoint requests. That mean if you want
            send some extra headers or data you can do it by additional kwargs.

        """

        results_id = []
        counter = 0
        for dt in data:
            result = self.get_result("post", url, 201, data=json.dumps(dt), **kwargs)
            self.assertTrue(isinstance(result, dict))
            for field in fields:
                st = "[~~ENCRYPTED~~]"
                if field == "vars" and st in result['vars'].values():
                    pass  # nocv
                else:
                    self.assertEqual(result[field], data[counter][field])
            results_id.append(result["id"])
            counter += 1
        return results_id

    def list_test(self, url, count):
        """
        Test for get list of models. Checks only list count. Uses :meth:`.get_result` method.

        :param url: url to abstract layer. For example: ``/api/v1/project/``.
                    You can use :meth:`.get_url` for building url.
        :param count: count of objects in DB.

        """
        result = self.get_result("get", url)
        self.assertTrue(isinstance(result, dict))
        self.assertEqual(result["count"], count)

    def details_test(self, url, **kwargs):
        """
        Test for get details of model. If you setup additional named arguments,
        the method check their equality with response data.
        Uses :meth:`.get_result` method.

        :param url: url to detail record. For example: ``/api/v1/project/1/`` (where ``1`` is uniq id of project).
                    You can use :meth:`.get_url` for building url.
        :param kwargs: params that's should be checked (key - field name, value - field value).

        """
        result = self.get_result("get", url)
        self.assertTrue(isinstance(result, dict))
        for key, value in kwargs.items():
            self.assertEqual(result[key], value)

    def endpoint_call(self, data: BulkDataType = None, method: str = 'get', code: int = 200, **kwargs) -> ApiResultType:
        """
        Make request to endpoint and assert response status code if specified (default is 200).
        Uses :meth:`.get_result` method for execution.

        :param data: request data
        :param method: http request method
        :param code: http status to assert
        :param query: dict with query data (working only with `get`)
        :return: bulk response
        """

        if data is not None:
            if self.use_msgpack:
                data = ormsgpack.packb(data)
            else:
                data = json.dumps(data)

        if method == 'get' and 'query' in kwargs and kwargs['query']:
            query = f'?{"&".join(map("=".join, kwargs.pop("query").items()))}'
        else:
            query = ''

        return self.get_result(
            method,
            f'/{self._settings("VST_API_URL")}/endpoint/{query}',
            data=data,
            code=code,
            HTTP_ACCEPT_ENCODING='gzip',
            **kwargs
        )

    def endpoint_schema(self, **kwargs):
        """
        Make request to schema. Returns dict with swagger data.

        :param version: API version for schema parser.
        """
        return self.endpoint_call(query={**kwargs, 'format': 'openapi'})

    def bulk(self, data: BulkDataType, code: int = 200, **kwargs) -> ApiResultType:
        """
        Make non transactional bulk request and assert status code (default is 200)

        :param data: request data
        :param code: http status to assert
        :param kwargs: named arguments for :meth:`.get_result`
        :return: bulk response
        """
        return self.endpoint_call(data, method='put', code=code, **kwargs)

    def bulk_transactional(self, data: BulkDataType, code: int = 200, **kwargs) -> ApiResultType:
        """
        Make transactional bulk request and assert status code (default is 200)

        :param data: request data
        :param code: http status to assert
        :param kwargs: named arguments for :meth:`.get_result`
        :return: bulk response
        """
        return self.endpoint_call(data, method='post', code=code, **kwargs)



# pylint: disable=no-member

import os
import sys
import time
from pathlib import Path
from subprocess import CalledProcessError
import subprocess
import signal

from django.conf import settings

from ._base import BaseCommand
from ...utils import raise_context

python_exec_dir = Path(os.path.dirname(settings.PYTHON_INTERPRETER))
python_subexec_dir = Path('/usr/local/bin')
_uwsgi_default_path = python_exec_dir / 'uwsgi'
_uwsgi_default_path_alt = python_exec_dir / 'pyuwsgi'
_uwsgi_default_path_alt2 = python_subexec_dir / 'uwsgi'
_uwsgi_default_path_alt3 = python_subexec_dir / 'pyuwsgi'
if not _uwsgi_default_path.exists() and _uwsgi_default_path_alt.exists():
    _uwsgi_default_path = _uwsgi_default_path_alt
elif _uwsgi_default_path_alt2.exists():
    _uwsgi_default_path = _uwsgi_default_path_alt2
elif _uwsgi_default_path_alt3.exists():
    _uwsgi_default_path = _uwsgi_default_path_alt3


def wait(proc, timeout=None, delay=0.01):
    while proc.poll() is None and (timeout or timeout is None):
        time.sleep(delay)
        if timeout is not None:
            timeout -= delay
    return proc.poll()


def get_celery_command(celery_path='celery', **kwargs):
    # Format args string.
    options = ''
    app_option = f'--app={settings.VST_PROJECT}.wapp:app'
    worker_options_dict = {
        **settings.WORKER_OPTIONS,
        **kwargs
    }
    for key, value in worker_options_dict.items():
        if key == 'app':
            app_option = "--app={}".format(value.replace(',', r'\,'))
            continue
        is_boolean = isinstance(value, bool)
        if (is_boolean and value) or value:
            options += f' -{"-" if key != "O" else ""}{key}'
        if is_boolean:
            continue
        options += "{}{}".format('=' if key != 'O' else '', value.replace(',', r'\,'))

    # Add queues list to celery args
    if '--queues' not in options:
        options += ' --queues={}'.format(r'\,'.join(settings.WORKER_QUEUES))

    # Add arguments to uwsgi cmd list.
    return f'{celery_path} {app_option} worker {options}'


class Command(BaseCommand):
    help = "Backend web-server."
    _uwsgi_default_path = _uwsgi_default_path
    uwsgi_default_config = Path(os.path.dirname(__file__)).parent.parent / 'web.ini'
    default_addrport = settings.WEB_ADDRPORT
    prefix = getattr(settings, 'VST_PROJECT_LIB', 'vstutils').upper()

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            'args',
            metavar='uwsgi_arg=value', nargs='*',
            help='Args "name=value" uwsgi server.',
        )
        parser.add_argument(
            '--uwsgi-path', '-s',
            default=self._uwsgi_default_path, dest='script',
            help='Specifies the uwsgi script.',
        )
        parser.add_argument(
            '--uwsgi-config', '-c',
            default=str(Path(settings.VST_PROJECT_DIR) / 'web.ini'),
            dest='config', help='Specifies the uwsgi script.',
        )
        parser.add_argument(
            '--addrport', '-p',
            default=self.default_addrport,
            dest='addrport', help='Specifies the uwsgi address:port. Default: [:8080]',
        )

    def _get_uwsgi_arg(self, arg):
        return arg if isinstance(arg, str) else None

    def _get_uwsgi_args(self, *uwsgi_args):
        args = []

        # Set `--daemonize` to logfile if `daemon = true`
        if settings.WEB_DAEMON:
            args.append(f'daemonize={settings.WEB_DAEMON_LOGFILE}')

        # Parse command args and setup to uwsgi
        args += [self._get_uwsgi_arg(arg) for arg in uwsgi_args]

        return args

    def _get_worker_options(self):
        cmd = []

        # Check that worker is enabled in settings.
        if not getattr(settings, 'RUN_WORKER', False):
            return cmd

        # Add arguments to uwsgi cmd list.
        cmd.append('--attach-daemon2')
        cmd.append(f'stopsignal=15,reloadsignal=1,exec={get_celery_command()}')

        return cmd

    def handle(self, *uwsgi_args, **opts):
        # pylint: disable=too-many-branches
        super().handle(*uwsgi_args, **opts)

        # Environment
        env = os.environ.copy()

        # Build default uwsgi-command options.
        cmd = [
            str(opts['script']),
            f'--set-ph=program_name={settings.VST_PROJECT}',
            f'--set-ph=lib_name={settings.VST_PROJECT_LIB}',
            f'--set-ph=api_path={settings.API_URL}',
            f'--set-ph=vstutils_version={settings.VSTUTILS_VERSION}',
            f'--module={settings.UWSGI_APPLICATION}',
        ]
        #  Setup http addr:port.
        addrport = opts['addrport']
        cmd += [f'--{"https" if len(addrport.split(",")) > 1 else "http"}={addrport}']

        # Import uwsgi-args from this command args (key=value).
        cmd += [
            f'--{arg}' for arg in self._get_uwsgi_args(*uwsgi_args)
            if arg is not None
        ]
        # Import config from project.
        if opts['config'] and Path(opts['config']).exists():
            cmd += [f'--ini={opts["config"]}']

        # Connect static files.
        if settings.STATIC_URL.startswith('/'):
            cmd += [f'--set-ph=static_url={settings.STATIC_URL}']
            for static_path in settings.STATIC_FILES_FOLDERS:
                if f"{settings.STATIC_URL}={static_path}" not in cmd:
                    cmd += ['--static-map', f"{settings.STATIC_URL}={static_path}"]

        if settings.MEDIA_URL.startswith('/') and settings.MEDIA_ROOT:
            cmd += ['--static-map', f"{settings.MEDIA_URL}={settings.MEDIA_ROOT}"]

        if 'docs' in settings.INSTALLED_APPS and getattr(settings, 'DOCS_ACCESS', 'public') == 'public':
            cmd += ['--static-map', f"{settings.DOC_URL}={settings.DOCS_ROOT}"]

        # Append uwsgi configs.
        for config_file in map(Path, settings.CONFIG_FILES):
            option = f'--ini={str(config_file)}'
            if config_file.exists() and config_file.suffix == '.ini' and option not in cmd:
                cmd.append(option)

        if self.uwsgi_default_config.exists():
            cmd.append(str(self.uwsgi_default_config))
        else:
            self._print(f'File {str(self.uwsgi_default_config)} doesnt exists.', 'ERROR')

        # Attach worker.
        cmd += self._get_worker_options()

        # Load config data from stdin.
        cmd += ['--ini', '-']

        # Check if it is run under virtualenv
        if sys.prefix != '/usr':
            cmd += [f'--virtualenv={sys.prefix}']
            if sys.prefix not in env["PATH"]:
                env['PATH'] = f'{sys.prefix}/bin:{env["PATH"]}'

        # Get config from env
        read, write = os.pipe()
        os.write(write, os.environ.get(self._settings('CONFIG_ENV_DATA_NAME'), '').encode('utf-8'))
        os.close(write)

        # Run web server
        try:
            self._print('Execute: ' + ' '.join(cmd))
            with subprocess.Popen(cmd, env=env, stdin=read) as proc:
                try:
                    wait(proc)
                except BaseException as exc:
                    proc.send_signal(signal.SIGTERM)
                    wait(proc, 10)
                    with raise_context():
                        proc.kill()
                    wait(proc)
                    raise exc
        except KeyboardInterrupt:
            self._print('Exit by user...', 'WARNING')
        except CalledProcessError as err:
            raise self.CommandError(str(err))
        except BaseException as err:
            self._print(str(err), 'ERROR')
            raise



import typing as _t
import operator
from functools import reduce

from django.conf import settings
from django.db.models import Q
from django.utils.functional import SimpleLazyObject
from django_filters import rest_framework as filters
from django_filters import CharFilter
from vstutils.utils import lazy_translate as __

from .. import utils

id_help = __('A unique integer value (or comma separated list) identifying this instance.')
name_help = __('A name string value (or comma separated list) of instance.')
name_filter_method = "contains" if settings.CASE_SENSITIVE_API_FILTER else 'icontains'


def _extra_query_search(field, value, stype):
    vals = field.rsplit("__", 1)
    field_name, tp = vals[0], (list(vals)[1:2] + [""])[0]
    negate = tp.upper() == "NOT"
    if negate:
        field = field_name
    field += f"__{stype}"
    value = value.split(",") if stype == "in" else value
    return {field: value}, negate


def _extra_search(queryset, field, value, stype):
    query_filter, negate = _extra_query_search(field, value, stype)
    if negate:
        return queryset.exclude(**query_filter)
    return queryset.filter(**query_filter)


def extra_filter(queryset, field, value):
    """
    Method for searching values in a comma-separated list.

    :param queryset: model queryset for filtration.
    :type queryset: django.db.models.query.QuerySet
    :param field: field name in FilterSet. Also supports `__not` suffix.
    :type field: str
    :param value: comma separated list of searching values.
    :type value: str
    :return: filtered queryset.
    :rtype: django.db.models.query.QuerySet
    """

    return _extra_search(queryset, field, value, "in")


def name_filter(queryset, field, value):
    """
    Method for searching by part of name. Uses `LIKE` DB condition or `contains` qs-expression.

    :param queryset: model queryset for filtration.
    :type queryset: django.db.models.query.QuerySet
    :param field: field name in FilterSet. Also supports `__not` suffix.
    :type field: str
    :param value: searching part of name.
    :type value: str
    :return: filtered queryset.
    :rtype: django.db.models.query.QuerySet
    """
    return _extra_search(queryset, field, value, name_filter_method)


class FkFilterHandler:
    """
    Simple handler for filtering by relational fields.

    :param related_pk: Field name of related model's primary key. Default is 'id'.
    :param related_name: Field name of related model's charfield. Default is 'name'.
    :param pk_handler: Changes handler for checking value before search. Sends "0" in handler falls. Default is 'int()'.

    Example:
        .. sourcecode:: python

            class CustomFilterSet(filters.FilterSet):
                author = CharFilter(method=vst_filters.FkFilterHandler(related_pk='pk', related_name='email'))

        Where ``author`` is ForeignKey to `User` and you want to search by primary key and email.
    """
    __slots__ = ('related_pk', 'related_name', 'pk_handler')

    def __init__(self, related_pk: str = 'id', related_name: str = 'name', pk_handler: _t.Callable = int):
        self.related_pk = related_pk
        self.related_name = related_name
        self.pk_handler = pk_handler

    def __get_q(self, kw, negate):
        query = Q(**kw)
        if negate:
            query.negate()
        return query

    def list_of_q(self, field, value):
        field_split = field.rsplit("__", 1)
        if len(field_split) > 1 and field_split[-1] == 'not':
            field, suffix = field_split[0], '__not'
        else:
            suffix = ''
        if self.related_pk:
            pk_value = SimpleLazyObject(
                utils.raise_context_decorator_with_default(default='0')(
                    lambda: str(self.pk_handler(value)) or "0"
                )
            )
            yield self.__get_q(*_extra_query_search(f'{field}__{self.related_pk}{suffix}', pk_value, "in"))
        if self.related_name:
            yield self.__get_q(*_extra_query_search(f'{field}__{self.related_name}{suffix}', value, name_filter_method))

    def __call__(self, queryset, field, value):
        return queryset.filter(
            reduce(
                operator.or_,
                self.list_of_q(field, value)
            )
        )


class DefaultIDFilter(filters.FilterSet):
    """
    Basic filterset to search by id. Provides a search for multiple values separated by commas.
    Uses :func:`.extra_filter` in fields.
    """
    id = CharFilter(method=extra_filter, help_text=id_help, label="Primary keys")
    id__not = CharFilter(method=extra_filter, help_text=id_help, label="Exclude primary keys")


class DefaultNameFilter(filters.FilterSet):
    """
    Basic filterset to search by part of name. Uses `LIKE` DB condition by :func:`.name_filter`.
    """
    name = CharFilter(method=name_filter, help_text=name_help, label="Name")
    name__not = CharFilter(method=name_filter, help_text=name_help, label='Exclude name')



from rest_framework import status, response


__globals = globals()


def __status_name_filter(name):
    return name.startswith('HTTP_')


class BaseResponseClass(response.Response):
    """
    API response class with default status code.

    :var status_code: HTTP status code.
    :vartype status_code: int

    :ivar timings: Response timings.
    :vartype timings: int,None

    :param timings: Response timings.
    :vartype timings: int,None
    """

    __slots__ = ('data', 'timings')

    def __init__(self, *args, **kwargs):
        self.timings = kwargs.pop('timings', None)
        super().__init__(*args, **kwargs)
        if isinstance(self.data, str):
            self.data = {'detail': self.data}


for __status_name in filter(__status_name_filter, dir(status)):
    __status_code = getattr(status, __status_name)
    __response_name = f'Response{__status_code}'
    __http_response_name = __status_name
    __globals[__response_name] = type(
        __response_name,
        (BaseResponseClass,),
        {"status_code": __status_code, "__slots__": ()}
    )
    __globals[__http_response_name] = __globals[__response_name]
